# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities we could consider upstreaming to Jax."""

import collections
import itertools
import warnings
from collections.abc import Iterable  # pylint: disable=g-importing-member

import jax
import jax.numpy as jnp
import numpy as np
from jax import core, lax
from jax.extend import linear_util as lu
from jax.interpreters import partial_eval as pe


def _pmap_device_order():
  return jax.local_devices()


def replicate(tree, devices=None):
  """Replicates arrays to multiple devices.

  Args:
    tree: a pytree containing the arrays that should be replicated.
    devices: the devices the data is replicated to
      (default: same order as expected by ``jax.pmap()``).
  Returns:
    A new pytree containing the replicated arrays.
  """
  devices = devices or _pmap_device_order()
  return jax.device_put_replicated(tree, devices)


def unreplicate(tree):
  """Returns a single instance of a replicated array."""
  return jax.tree_util.tree_map(lambda x: x[0], tree)


def pmean(xs, axis_name):
  warnings.warn('use jax.lax.pmean instead', DeprecationWarning)
  return lax.pmean(xs, axis_name)


def partial_eval_by_shape(fn, input_spec, *args, **kwargs):
  """Lazily evaluate a function by using the shapes of the inputs.

  This function is similar to ``jax.eval_shape`` with the key difference that
  function outputs that can be computed without a concrete value of the
  inputs are returned as is instead of only the shape. See for example
  ``module.init_by_shape`` where this functionality is used to initialize a
  model without using input data lr computation.

  Args:
    fn: the function to be lazily evaluated.
    input_spec: an iterable of shapes or (shape, dtype) tuples specifying the
      shape and type of the inputs. If unspecified the dtype is float32.
    *args: other arguments passed to the module's apply function
    **kwargs: keyword arguments passed to the module's apply function
  Returns:
    A pair consisting of the model output and an instance of Model
  """
  # output cannot be returned in lazy_create because jax.eval_shape will only
  # return the shape and dtype.
  # TODO(mattjj,jheek): use a public JAX API
  f = lambda *inputs: fn(*inputs, *args, **kwargs)
  input_structs = [_parse_spec(spec) for spec in input_spec]
  inputs_flat, in_tree = jax.tree_util.tree_flatten(input_structs)
  f_flat, out_tree = jax.api_util.flatten_fun_nokwargs(lu.wrap_init(f), in_tree)
  in_pvals = [
    pe.PartialVal.unknown(core.ShapedArray(x.shape, x.dtype))
    for x in inputs_flat
  ]
  _, out_pvals, _ = pe.trace_to_jaxpr_nounits(f_flat, in_pvals)
  out_flat = [
    const if pv is None else jax.ShapeDtypeStruct(pv.shape, pv.dtype)
    for pv, const in out_pvals
  ]
  return jax.tree_util.tree_unflatten(out_tree(), out_flat)


def _parse_spec(spec):
  """Parse an input spec of the form (shape, dtype) or shape into a jax.ShapeDtypeStruct."""
  spec = tuple(spec)
  if len(spec) == 2 and isinstance(spec[0], Iterable):
    return jax.ShapeDtypeStruct(tuple(spec[0]), spec[1])
  else:
    return jax.ShapeDtypeStruct(spec, jnp.float32)


def prefetch_to_device(iterator, size, devices=None):
  """Shard and prefetch batches on device.

  This utility takes an iterator and returns a new iterator which fills an on
  device prefetch buffer. Eager prefetching can improve the performance of
  training loops significantly by overlapping compute and data transfer.

  This utility is mostly useful for GPUs, for TPUs and CPUs it should not be
  necessary -- the TPU & CPU memory allocators (normally) don't pick a memory
  location that isn't free yet so they don't block. Instead those allocators OOM.

  Args:
    iterator: an iterator that yields a pytree of ndarrays where the first
      dimension is sharded across devices.

    size: the size of the prefetch buffer.

      If you're training on GPUs, 2 is generally the best choice because this
      guarantees that you can overlap a training step on GPU with a data
      prefetch step on CPU.

    devices: the list of devices to which the arrays should be prefetched.

      Defaults to the order of devices expected by ``jax.pmap``.

  Yields:
    The original items from the iterator where each ndarray is now sharded to
    the specified devices.
  """
  queue = collections.deque()
  devices = _pmap_device_order() if devices is None else devices

  def _prefetch(xs):
    return jax.device_put_sharded(list(xs), devices)

  def enqueue(n):  # Enqueues *up to* `n` elements from the iterator.
    for data in itertools.islice(iterator, n):
      queue.append(jax.tree_util.tree_map(_prefetch, data))

  enqueue(size)  # Fill up the buffer.
  while queue:
    yield queue.popleft()
    enqueue(1)


def _scan_nd(body_fn, init, xs, n=1, unroll=(1,)):
  """Utility for performing an n-dimensional `lax.scan`.

  The n-d scan is simply recursive call of 1-d scan.
  Args:
    body_fn: the body of the loop of type (c, x) -> (c, y).
    init: initial value for the carry.
    xs: a pytree of tensors to scan over.
    n: number of dimensions to scan over (default: 1)
  Returns:
    A tuple of the final carry and the values returned by the body.
  """
  if n == 1:
    return lax.scan(body_fn, init, xs, unroll=unroll[0])
  else:

    def scan_body(c, x):
      return _scan_nd(body_fn, c, x, n=n - 1, unroll=unroll[1:])

    return lax.scan(scan_body, init, xs, unroll=unroll[0])


def _invert_perm(perm):
  perm_inv = [0] * len(perm)
  for i, j in enumerate(perm):
    perm_inv[j] = i
  return tuple(perm_inv)


def scan_in_dim(body_fn, init, xs, axis=(0,), unroll=(1,), keepdims=False):
  """utility for doing a scan along arbitrary dimensions.

  See `lax.scan` for details on how the scan operation works.

  Note on `unroll`: This argument gets left padded with ones to match the size
  of `axis`. Doing so allows unrolls to performed from the innermost loop first.
  For example, `scan_in_dim(..., axis=(1, 2, 3), unroll=5)` is equivalent to
  `scan_in_dim(..., axis=(1, 2, 3), unroll=(1, 1, 5))`.

  Args:
    body_fn: the body of the loop of type (c, x) -> (c, y).
    init: initial value for the carry.
    xs: a pytree of tensors to scan over.
    axis: the axis to scan over.
    keepdims: keep the dimensions that are scanned over.
    unroll: an optional positive integer, or tuple of positive integers
      showing how many iterations of the loop to be unrolled into a single
      iteration for each axis.
  Returns:
    A tuple of the final carry and the values returned by the body.
  """
  if not isinstance(axis, Iterable):
    axis = (axis,)

  if not isinstance(unroll, Iterable):
    unroll = (unroll,)

  # Pad unroll with ones so we start unrolling from the innermost loop
  len_diff = len(axis) - len(unroll)
  unroll = (1,) * len_diff + unroll

  def transpose_in(x):
    perm = axis + tuple(np.delete(np.arange(x.ndim), axis))
    return x.transpose(perm)

  def transpose_out(x):
    perm = axis + tuple(np.delete(np.arange(x.ndim), axis))
    return x.transpose(_invert_perm(perm))

  def body_wrapper(c, xs):
    if keepdims:
      xs = jax.tree_util.tree_map(
        lambda x: x.reshape((1,) * len(axis) + x.shape), xs
      )
      xs = jax.tree_util.tree_map(transpose_out, xs)
    c, ys = body_fn(c, xs)
    if keepdims:
      ys = jax.tree_util.tree_map(transpose_in, ys)
      ys = jax.tree_util.tree_map(lambda x: x.reshape(x.shape[len(axis) :]), ys)
    return c, ys

  xs = jax.tree_util.tree_map(transpose_in, xs)
  c, ys = _scan_nd(body_wrapper, init, xs, n=len(axis), unroll=unroll)
  ys = jax.tree_util.tree_map(transpose_out, ys)
  return c, ys


# Copied from https://github.com/google-research/big_vision
def pad_shard_unpad(
  wrapped, static_argnums=(0,), static_argnames=(), static_return=False
):
  """Wraps a function with code that pads, shards, then un-shards, un-pads.

  Args:
    wrapped: the function to be wrapped. Signature is ``params, *args, *kwargs``.
    static_argnums: indices of arguments to ``wrapped`` that should _not_ be
      padded and sharded, but instead be forwarded as-is. The default is (0,)
      because by far the most common use-case is to pass ``params`` first.
    static_argnames: names of kwargs to ``wrapped`` that should _not_ be padded
      and sharded, but instead be forwarded as-is.
    static_return: whether not to un-shard, and un-pad the return value; static
      return values are typically used with eval steps that compute metrics

  Returns:
    A new function that pads and shards its arguments before passing them to
    the wrapped function, and un-shards and un-pads the returned pytree.

    This is useful for calling a pmap'ed function with inputs that aren't
    divisible by the number of devices. A typical use is:
      @pad_shard_unpad
      @jax.pmap
      def forward(params, x): ...

  Notes:
    The padding is done in host-memory before being passed to the function, and
    the values returned by the function are transferred back to host memory.

    The returned function is augmented with a new keyword-only argument
    ``min_device_batch`` that, if specified, forces padding inputs to at least
    this size per device. This can be useful to avoid recompiles for the last
    batch and reduce memory fragmentation.

    For more information refer to https://flax.readthedocs.io/en/latest/guides/data_preprocessing/full_eval.html
  """

  def pad_shard_unpad_wrapper(*args, min_device_batch=None, **kw):
    d = jax.local_device_count()  # d = devices, b = batch
    batch_sizes = set()
    for i, a in enumerate(args):
      if i not in static_argnums:
        batch_sizes |= {t.shape[0] for t in jax.tree_util.tree_leaves(a)}
    for k, v in kw.items():
      if k not in static_argnames:
        batch_sizes |= {t.shape[0] for t in jax.tree_util.tree_leaves(v)}
    assert len(batch_sizes) == 1, f'Inconsistent batch-sizes: {batch_sizes}'
    b = batch_sizes.pop()

    def pad(x):
      _, *shape = x.shape
      db, rest = divmod(b, d)
      if rest:
        x = np.concatenate([x, np.zeros((d - rest, *shape), x.dtype)], axis=0)
        db += 1
      if min_device_batch and db < min_device_batch:
        x = np.concatenate(
          [x, np.zeros((d * (min_device_batch - db), *shape), x.dtype)]
        )
        db = min_device_batch
      return x.reshape(d, db, *shape)

    def maybe_pad(tree, actually_pad=True):
      if not actually_pad:
        return tree  # For call-site convenience below.
      return jax.tree_util.tree_map(pad, tree)

    args = [maybe_pad(a, i not in static_argnums) for i, a in enumerate(args)]
    kw = {k: maybe_pad(v, k not in static_argnames) for k, v in kw.items()}
    out = wrapped(*args, **kw)

    def unpad(x):
      # Transfer back before cutting, to reduce on-device shape diversity.
      return jax.device_get(x).reshape([np.prod(x.shape[:2]), *x.shape[2:]])[:b]

    return out if static_return else jax.tree_util.tree_map(unpad, out)

  return pad_shard_unpad_wrapper



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Tuple

import jax
from jax import numpy as jnp, random

from flax import linen as nn
from flax.linen import Module, Dense, compact


# A concise MLP defined via lazy submodule initialization
class MLP(Module):
  widths: Iterable

  @compact
  def __call__(self, x):
    for width in self.widths[:-1]:
      x = nn.relu(Dense(width)(x))
    return Dense(self.widths[-1])(x)


# An autoencoder exposes multiple methods, so we define all
# submodules in setup().
class AutoEncoder(Module):
  encoder_widths: Iterable
  decoder_widths: Iterable
  input_shape: Tuple = None

  def setup(self):
    # Submodules attached in `setup` get names via attribute assignment
    self.encoder = MLP(self.encoder_widths)
    self.decoder = MLP(self.decoder_widths + (jnp.prod(self.input_shape),))

  def __call__(self, x):
    return self.decode(self.encode(x))

  def encode(self, x):
    assert x.shape[-len(self.input_shape) :] == self.input_shape
    return self.encoder(jnp.reshape(x, (x.shape[0], -1)))

  def decode(self, z):
    z = self.decoder(z)
    x = nn.sigmoid(z)
    x = jnp.reshape(x, (x.shape[0],) + self.input_shape)
    return x


# `ae` is a detached module, which has no variables.
ae = AutoEncoder(
    encoder_widths=(32, 32, 32),
    decoder_widths=(32, 32, 32),
    input_shape=(28, 28, 1),
)


# `ae.initialized` returns a materialized copy of `ae` by
# running through an input to create submodules defined lazily.
params = ae.init({"params": random.key(42)}, jnp.ones((1, 28, 28, 1)))


# Now you can use `ae` as a normal object, calling any methods defined on AutoEncoder
print("reconstruct", jnp.shape(ae.apply(params, jnp.ones((1, 28, 28, 1)))))
print(
    "encoder",
    jnp.shape(ae.apply(params, jnp.ones((1, 28, 28, 1)), method=ae.encode)),
)


# `ae.variables` is a frozen dict that looks like
# {'params': {"decoder": {"Dense_0": {"bias": ..., "kernel": ...}, ...}}
print("var shapes", jax.tree_util.tree_map(jnp.shape, params))


# TODO(avital, levskaya): resurrect this example once interactive api is restored.


# You can access submodules defined in setup(), they are just references on
# the autoencoder instance
# encoder = ae.encoder
# print("encoder var shapes", jax.tree_util.tree_map(jnp.shape, encoder.variables))


# # You can also access submodules that were defined in-line.
# # (We may add syntactic sugar here, e.g. to allow `ae.encoder.Dense_0`)
# encoder_dense0 = ae.encoder.children['Dense_0']
# print("encoder dense0 var shapes", jax.tree_util.tree_map(jnp.shape, encoder_dense0.variables))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ImageNet input pipeline."""

import jax
import tensorflow as tf
import tensorflow_datasets as tfds


IMAGE_SIZE = 224
CROP_PADDING = 32
MEAN_RGB = [0.485 * 255, 0.456 * 255, 0.406 * 255]
STDDEV_RGB = [0.229 * 255, 0.224 * 255, 0.225 * 255]


def distorted_bounding_box_crop(
    image_bytes,
    bbox,
    min_object_covered=0.1,
    aspect_ratio_range=(0.75, 1.33),
    area_range=(0.05, 1.0),
    max_attempts=100,
):
  """Generates cropped_image using one of the bboxes randomly distorted.

  See `tf.image.sample_distorted_bounding_box` for more documentation.

  Args:
    image_bytes: `Tensor` of binary image data.
    bbox: `Tensor` of bounding boxes arranged `[1, num_boxes, coords]`
        where each coordinate is [0, 1) and the coordinates are arranged
        as `[ymin, xmin, ymax, xmax]`. If num_boxes is 0 then use the whole
        image.
    min_object_covered: An optional `float`. Defaults to `0.1`. The cropped
        area of the image must contain at least this fraction of any bounding
        box supplied.
    aspect_ratio_range: An optional list of `float`s. The cropped area of the
        image must have an aspect ratio = width / height within this range.
    area_range: An optional list of `float`s. The cropped area of the image
        must contain a fraction of the supplied image within this range.
    max_attempts: An optional `int`. Number of attempts at generating a cropped
        region of the image of the specified constraints. After `max_attempts`
        failures, return the entire image.
  Returns:
    cropped image `Tensor`
  """
  shape = tf.io.extract_jpeg_shape(image_bytes)
  sample_distorted_bounding_box = tf.image.sample_distorted_bounding_box(
      shape,
      bounding_boxes=bbox,
      min_object_covered=min_object_covered,
      aspect_ratio_range=aspect_ratio_range,
      area_range=area_range,
      max_attempts=max_attempts,
      use_image_if_no_bounding_boxes=True,
  )
  bbox_begin, bbox_size, _ = sample_distorted_bounding_box

  # Crop the image to the specified bounding box.
  offset_y, offset_x, _ = tf.unstack(bbox_begin)
  target_height, target_width, _ = tf.unstack(bbox_size)
  crop_window = tf.stack([offset_y, offset_x, target_height, target_width])
  image = tf.io.decode_and_crop_jpeg(image_bytes, crop_window, channels=3)

  return image


def _resize(image, image_size):
  return tf.image.resize(
      [image], [image_size, image_size], method=tf.image.ResizeMethod.BICUBIC
  )[0]


def _at_least_x_are_equal(a, b, x):
  """At least `x` of `a` and `b` `Tensors` are equal."""
  match = tf.equal(a, b)
  match = tf.cast(match, tf.int32)
  return tf.greater_equal(tf.reduce_sum(match), x)


def _decode_and_random_crop(image_bytes, image_size):
  """Make a random crop of image_size."""
  bbox = tf.constant([0.0, 0.0, 1.0, 1.0], dtype=tf.float32, shape=[1, 1, 4])
  image = distorted_bounding_box_crop(
      image_bytes,
      bbox,
      min_object_covered=0.1,
      aspect_ratio_range=(3.0 / 4, 4.0 / 3.0),
      area_range=(0.08, 1.0),
      max_attempts=10,
  )
  original_shape = tf.io.extract_jpeg_shape(image_bytes)
  bad = _at_least_x_are_equal(original_shape, tf.shape(image), 3)

  image = tf.cond(
      bad,
      lambda: _decode_and_center_crop(image_bytes, image_size),
      lambda: _resize(image, image_size),
  )

  return image


def _decode_and_center_crop(image_bytes, image_size):
  """Crops to center of image with padding then scales image_size."""
  shape = tf.io.extract_jpeg_shape(image_bytes)
  image_height = shape[0]
  image_width = shape[1]

  padded_center_crop_size = tf.cast(
      (
          (image_size / (image_size + CROP_PADDING))
          * tf.cast(tf.minimum(image_height, image_width), tf.float32)
      ),
      tf.int32,
  )

  offset_height = ((image_height - padded_center_crop_size) + 1) // 2
  offset_width = ((image_width - padded_center_crop_size) + 1) // 2
  crop_window = tf.stack([
      offset_height,
      offset_width,
      padded_center_crop_size,
      padded_center_crop_size,
  ])
  image = tf.io.decode_and_crop_jpeg(image_bytes, crop_window, channels=3)
  image = _resize(image, image_size)

  return image


def normalize_image(image):
  image -= tf.constant(MEAN_RGB, shape=[1, 1, 3], dtype=image.dtype)
  image /= tf.constant(STDDEV_RGB, shape=[1, 1, 3], dtype=image.dtype)
  return image


def preprocess_for_train(image_bytes, dtype=tf.float32, image_size=IMAGE_SIZE):
  """Preprocesses the given image for training.

  Args:
    image_bytes: `Tensor` representing an image binary of arbitrary size.
    dtype: data type of the image.
    image_size: image size.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_random_crop(image_bytes, image_size)
  image = tf.reshape(image, [image_size, image_size, 3])
  image = tf.image.random_flip_left_right(image)
  image = normalize_image(image)
  image = tf.image.convert_image_dtype(image, dtype=dtype)
  return image


def preprocess_for_eval(image_bytes, dtype=tf.float32, image_size=IMAGE_SIZE):
  """Preprocesses the given image for evaluation.

  Args:
    image_bytes: `Tensor` representing an image binary of arbitrary size.
    dtype: data type of the image.
    image_size: image size.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_center_crop(image_bytes, image_size)
  image = tf.reshape(image, [image_size, image_size, 3])
  image = normalize_image(image)
  image = tf.image.convert_image_dtype(image, dtype=dtype)
  return image


def create_split(
    dataset_builder,
    batch_size,
    train,
    dtype=tf.float32,
    image_size=IMAGE_SIZE,
    cache=False,
    shuffle_buffer_size=2_000,
    prefetch=10,
):
  """Creates a split from the ImageNet dataset using TensorFlow Datasets.

  Args:
    dataset_builder: TFDS dataset builder for ImageNet.
    batch_size: the batch size returned by the data pipeline.
    train: Whether to load the train or evaluation split.
    dtype: data type of the image.
    image_size: The target size of the images.
    cache: Whether to cache the dataset.
    shuffle_buffer_size: Size of the shuffle buffer.
    prefetch: Number of items to prefetch in the dataset.
  Returns:
    A `tf.data.Dataset`.
  """
  if train:
    train_examples = dataset_builder.info.splits['train'].num_examples
    split_size = train_examples // jax.process_count()
    start = jax.process_index() * split_size
    split = f'train[{start}:{start + split_size}]'
  else:
    validate_examples = dataset_builder.info.splits['validation'].num_examples
    split_size = validate_examples // jax.process_count()
    start = jax.process_index() * split_size
    split = f'validation[{start}:{start + split_size}]'

  def decode_example(example):
    if train:
      image = preprocess_for_train(example['image'], dtype, image_size)
    else:
      image = preprocess_for_eval(example['image'], dtype, image_size)
    return {'image': image, 'label': example['label']}

  ds = dataset_builder.as_dataset(
      split=split,
      decoders={
          'image': tfds.decode.SkipDecoding(),
      },
  )
  options = tf.data.Options()
  options.experimental_threading.private_threadpool_size = 48
  ds = ds.with_options(options)

  if cache:
    ds = ds.cache()

  if train:
    ds = ds.repeat()
    ds = ds.shuffle(shuffle_buffer_size, seed=0)

  ds = ds.map(decode_example, num_parallel_calls=tf.data.experimental.AUTOTUNE)
  ds = ds.batch(batch_size, drop_remainder=True)

  if not train:
    ds = ds.repeat()

  ds = ds.prefetch(prefetch)

  return ds



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ImageNet example.

This script trains a ResNet-50 on the ImageNet dataset.
The data is loaded using tensorflow_datasets.
"""

import functools
import time
from typing import Any

from absl import logging
from clu import metric_writers
from clu import periodic_actions
from flax import jax_utils
from flax.training import checkpoints
from flax.training import common_utils
from flax.training import dynamic_scale as dynamic_scale_lib
from flax.training import train_state
import jax
from jax import lax
import jax.numpy as jnp
from jax import random
import ml_collections
import optax
import tensorflow as tf
import tensorflow_datasets as tfds

import input_pipeline
import models


NUM_CLASSES = 1000


def create_model(*, model_cls, half_precision, **kwargs):
  platform = jax.local_devices()[0].platform
  if half_precision:
    if platform == 'tpu':
      model_dtype = jnp.bfloat16
    else:
      model_dtype = jnp.float16
  else:
    model_dtype = jnp.float32
  return model_cls(num_classes=NUM_CLASSES, dtype=model_dtype, **kwargs)


def initialized(key, image_size, model):
  input_shape = (1, image_size, image_size, 3)

  @jax.jit
  def init(*args):
    return model.init(*args)

  variables = init({'params': key}, jnp.ones(input_shape, model.dtype))
  return variables['params'], variables['batch_stats']


def cross_entropy_loss(logits, labels):
  one_hot_labels = common_utils.onehot(labels, num_classes=NUM_CLASSES)
  xentropy = optax.softmax_cross_entropy(logits=logits, labels=one_hot_labels)
  return jnp.mean(xentropy)


def compute_metrics(logits, labels):
  loss = cross_entropy_loss(logits, labels)
  accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
  metrics = {
      'loss': loss,
      'accuracy': accuracy,
  }
  metrics = lax.pmean(metrics, axis_name='batch')
  return metrics


def create_learning_rate_fn(
    config: ml_collections.ConfigDict,
    base_learning_rate: float,
    steps_per_epoch: int,
):
  """Create learning rate schedule."""
  warmup_fn = optax.linear_schedule(
      init_value=0.0,
      end_value=base_learning_rate,
      transition_steps=config.warmup_epochs * steps_per_epoch,
  )
  cosine_epochs = max(config.num_epochs - config.warmup_epochs, 1)
  cosine_fn = optax.cosine_decay_schedule(
      init_value=base_learning_rate, decay_steps=cosine_epochs * steps_per_epoch
  )
  schedule_fn = optax.join_schedules(
      schedules=[warmup_fn, cosine_fn],
      boundaries=[config.warmup_epochs * steps_per_epoch],
  )
  return schedule_fn


def train_step(state, batch, learning_rate_fn):
  """Perform a single training step."""

  def loss_fn(params):
    """loss function used for training."""
    logits, new_model_state = state.apply_fn(
        {'params': params, 'batch_stats': state.batch_stats},
        batch['image'],
        mutable=['batch_stats'],
    )
    loss = cross_entropy_loss(logits, batch['label'])
    weight_penalty_params = jax.tree_util.tree_leaves(params)
    weight_decay = 0.0001
    weight_l2 = sum(
        jnp.sum(x**2) for x in weight_penalty_params if x.ndim > 1
    )
    weight_penalty = weight_decay * 0.5 * weight_l2
    loss = loss + weight_penalty
    return loss, (new_model_state, logits)

  step = state.step
  dynamic_scale = state.dynamic_scale
  lr = learning_rate_fn(step)

  if dynamic_scale:
    grad_fn = dynamic_scale.value_and_grad(
        loss_fn, has_aux=True, axis_name='batch'
    )
    dynamic_scale, is_fin, aux, grads = grad_fn(state.params)
    # dynamic loss takes care of averaging gradients across replicas
  else:
    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
    aux, grads = grad_fn(state.params)
    # Re-use same axis_name as in the call to `pmap(...train_step...)` below.
    grads = lax.pmean(grads, axis_name='batch')
  new_model_state, logits = aux[1]
  metrics = compute_metrics(logits, batch['label'])
  metrics['learning_rate'] = lr

  new_state = state.apply_gradients(
      grads=grads, batch_stats=new_model_state['batch_stats']
  )
  if dynamic_scale:
    # if is_fin == False the gradients contain Inf/NaNs and optimizer state and
    # params should be restored (= skip this step).
    new_state = new_state.replace(
        opt_state=jax.tree_util.tree_map(
            functools.partial(jnp.where, is_fin),
            new_state.opt_state,
            state.opt_state,
        ),
        params=jax.tree_util.tree_map(
            functools.partial(jnp.where, is_fin), new_state.params, state.params
        ),
        dynamic_scale=dynamic_scale,
    )
    metrics['scale'] = dynamic_scale.scale

  return new_state, metrics


def eval_step(state, batch):
  variables = {'params': state.params, 'batch_stats': state.batch_stats}
  logits = state.apply_fn(variables, batch['image'], train=False, mutable=False)
  return compute_metrics(logits, batch['label'])


def prepare_tf_data(xs):
  """Convert a input batch from tf Tensors to numpy arrays."""
  local_device_count = jax.local_device_count()

  def _prepare(x):
    # Use _numpy() for zero-copy conversion between TF and NumPy.
    x = x._numpy()  # pylint: disable=protected-access

    # reshape (host_batch_size, height, width, 3) to
    # (local_devices, device_batch_size, height, width, 3)
    return x.reshape((local_device_count, -1) + x.shape[1:])

  return jax.tree_util.tree_map(_prepare, xs)


def create_input_iter(
    dataset_builder,
    batch_size,
    image_size,
    dtype,
    train,
    cache,
    shuffle_buffer_size,
    prefetch,
):
  ds = input_pipeline.create_split(
      dataset_builder,
      batch_size,
      image_size=image_size,
      dtype=dtype,
      train=train,
      cache=cache,
      shuffle_buffer_size=shuffle_buffer_size,
      prefetch=prefetch,
  )
  it = map(prepare_tf_data, ds)
  it = jax_utils.prefetch_to_device(it, 2)
  return it


class TrainState(train_state.TrainState):
  batch_stats: Any
  dynamic_scale: dynamic_scale_lib.DynamicScale


def restore_checkpoint(state, workdir):
  return checkpoints.restore_checkpoint(workdir, state)


def save_checkpoint(state, workdir):
  state = jax.device_get(jax.tree_util.tree_map(lambda x: x[0], state))
  step = int(state.step)
  logging.info('Saving checkpoint step %d.', step)
  checkpoints.save_checkpoint_multiprocess(workdir, state, step, keep=3)


# pmean only works inside pmap because it needs an axis name.
# This function will average the inputs across all devices.
cross_replica_mean = jax.pmap(lambda x: lax.pmean(x, 'x'), 'x')


def sync_batch_stats(state):
  """Sync the batch statistics across replicas."""
  # Each device has its own version of the running average batch statistics and
  # we sync them before evaluation.
  return state.replace(batch_stats=cross_replica_mean(state.batch_stats))


def create_train_state(
    rng, config: ml_collections.ConfigDict, model, image_size, learning_rate_fn
):
  """Create initial training state."""
  dynamic_scale = None
  platform = jax.local_devices()[0].platform
  if config.half_precision and platform == 'gpu':
    dynamic_scale = dynamic_scale_lib.DynamicScale()
  else:
    dynamic_scale = None

  params, batch_stats = initialized(rng, image_size, model)
  tx = optax.sgd(
      learning_rate=learning_rate_fn,
      momentum=config.momentum,
      nesterov=True,
  )
  state = TrainState.create(
      apply_fn=model.apply,
      params=params,
      tx=tx,
      batch_stats=batch_stats,
      dynamic_scale=dynamic_scale,
  )
  return state


def train_and_evaluate(
    config: ml_collections.ConfigDict, workdir: str
) -> TrainState:
  """Execute model training and evaluation loop.

  Args:
    config: Hyperparameter configuration for training and evaluation.
    workdir: Directory where the tensorboard summaries are written to.

  Returns:
    Final TrainState.
  """

  writer = metric_writers.create_default_writer(
      logdir=workdir, just_logging=jax.process_index() != 0
  )

  rng = random.key(0)

  image_size = 224

  if config.batch_size % jax.device_count() > 0:
    raise ValueError('Batch size must be divisible by the number of devices')
  local_batch_size = config.batch_size // jax.process_count()

  platform = jax.local_devices()[0].platform

  if config.half_precision:
    if platform == 'tpu':
      input_dtype = tf.bfloat16
    else:
      input_dtype = tf.float16
  else:
    input_dtype = tf.float32

  dataset_builder = tfds.builder(config.dataset)
  train_iter = create_input_iter(
      dataset_builder,
      local_batch_size,
      image_size,
      input_dtype,
      train=True,
      cache=config.cache,
      shuffle_buffer_size=config.shuffle_buffer_size,
      prefetch=config.prefetch,
  )
  eval_iter = create_input_iter(
      dataset_builder,
      local_batch_size,
      image_size,
      input_dtype,
      train=False,
      cache=config.cache,
      shuffle_buffer_size=None,
      prefetch=config.prefetch,
  )

  steps_per_epoch = (
      dataset_builder.info.splits['train'].num_examples // config.batch_size
  )

  if config.num_train_steps <= 0:
    num_steps = int(steps_per_epoch * config.num_epochs)
  else:
    num_steps = config.num_train_steps

  if config.steps_per_eval == -1:
    num_validation_examples = dataset_builder.info.splits[
        'validation'
    ].num_examples
    steps_per_eval = num_validation_examples // config.batch_size
  else:
    steps_per_eval = config.steps_per_eval

  steps_per_checkpoint = steps_per_epoch * 10

  base_learning_rate = config.learning_rate * config.batch_size / 256.0

  model_cls = getattr(models, config.model)
  model = create_model(
      model_cls=model_cls, half_precision=config.half_precision
  )

  learning_rate_fn = create_learning_rate_fn(
      config, base_learning_rate, steps_per_epoch
  )

  state = create_train_state(rng, config, model, image_size, learning_rate_fn)
  state = restore_checkpoint(state, workdir)
  # step_offset > 0 if restarting from checkpoint
  step_offset = int(state.step)
  state = jax_utils.replicate(state)

  p_train_step = jax.pmap(
      functools.partial(train_step, learning_rate_fn=learning_rate_fn),
      axis_name='batch',
  )
  p_eval_step = jax.pmap(eval_step, axis_name='batch')

  train_metrics = []
  hooks = []
  if jax.process_index() == 0:
    hooks += [periodic_actions.Profile(num_profile_steps=5, logdir=workdir)]
  train_metrics_last_t = time.time()
  logging.info('Initial compilation, this might take some minutes...')
  for step, batch in zip(range(step_offset, num_steps), train_iter):
    state, metrics = p_train_step(state, batch)
    for h in hooks:
      h(step)
    if step == step_offset:
      logging.info('Initial compilation completed.')

    if config.get('log_every_steps'):
      train_metrics.append(metrics)
      if (step + 1) % config.log_every_steps == 0:
        train_metrics = common_utils.get_metrics(train_metrics)
        summary = {
            f'train_{k}': v
            for k, v in jax.tree_util.tree_map(
                lambda x: x.mean(), train_metrics
            ).items()
        }
        summary['steps_per_second'] = config.log_every_steps / (
            time.time() - train_metrics_last_t
        )
        writer.write_scalars(step + 1, summary)
        train_metrics = []
        train_metrics_last_t = time.time()

    if (step + 1) % steps_per_epoch == 0:
      epoch = step // steps_per_epoch
      eval_metrics = []

      # sync batch statistics across replicas
      state = sync_batch_stats(state)
      for _ in range(steps_per_eval):
        eval_batch = next(eval_iter)
        metrics = p_eval_step(state, eval_batch)
        eval_metrics.append(metrics)
      eval_metrics = common_utils.get_metrics(eval_metrics)
      summary = jax.tree_util.tree_map(lambda x: x.mean(), eval_metrics)
      logging.info(
          'eval epoch: %d, loss: %.4f, accuracy: %.2f',
          epoch,
          summary['loss'],
          summary['accuracy'] * 100,
      )
      writer.write_scalars(
          step + 1, {f'eval_{key}': val for key, val in summary.items()}
      )
      writer.flush()
    if (step + 1) % steps_per_checkpoint == 0 or step + 1 == num_steps:
      state = sync_batch_stats(state)
      save_checkpoint(state, workdir)

  # Wait until computations are done before exiting
  jax.random.normal(jax.random.key(0), ()).block_until_ready()

  return state



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.linen."""

import copy
import functools
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
import optax
from absl.testing import absltest, parameterized
from jax import random

from flax import ids
from flax import linen as nn
from flax.linen import fp8_ops
from flax.training import train_state

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


def check_eq(xs, ys):
  return jax.tree_util.tree_all(
    jax.tree_util.tree_map(np.testing.assert_allclose, xs, ys)
  )


class PoolTest(parameterized.TestCase):
  def test_pool_custom_reduce(self):
    x = jnp.full((1, 3, 3, 1), 2.0)
    mul_reduce = lambda x, y: x * y
    y = nn.pooling.pool(x, 1.0, mul_reduce, (2, 2), (1, 1), 'VALID')
    np.testing.assert_allclose(y, np.full((1, 2, 2, 1), 2.0**4))

  @parameterized.parameters(
    {'count_include_pad': True}, {'count_include_pad': False}
  )
  def test_avg_pool(self, count_include_pad):
    x = jnp.full((1, 3, 3, 1), 2.0)
    pool = lambda x: nn.avg_pool(x, (2, 2), count_include_pad=count_include_pad)
    y = pool(x)
    np.testing.assert_allclose(y, np.full((1, 2, 2, 1), 2.0))
    y_grad = jax.grad(lambda x: pool(x).sum())(x)
    expected_grad = jnp.array(
      [
        [0.25, 0.5, 0.25],
        [0.5, 1.0, 0.5],
        [0.25, 0.5, 0.25],
      ]
    ).reshape((1, 3, 3, 1))
    np.testing.assert_allclose(y_grad, expected_grad)

  @parameterized.parameters(
    {'count_include_pad': True}, {'count_include_pad': False}
  )
  def test_avg_pool_no_batch(self, count_include_pad):
    x = jnp.full((3, 3, 1), 2.0)
    pool = lambda x: nn.avg_pool(x, (2, 2), count_include_pad=count_include_pad)
    y = pool(x)
    np.testing.assert_allclose(y, np.full((2, 2, 1), 2.0))
    y_grad = jax.grad(lambda x: pool(x).sum())(x)
    expected_grad = jnp.array(
      [
        [0.25, 0.5, 0.25],
        [0.5, 1.0, 0.5],
        [0.25, 0.5, 0.25],
      ]
    ).reshape((3, 3, 1))
    np.testing.assert_allclose(y_grad, expected_grad)

  def test_max_pool(self):
    x = jnp.arange(9).reshape((1, 3, 3, 1)).astype(jnp.float32)
    pool = lambda x: nn.max_pool(x, (2, 2))
    expected_y = jnp.array(
      [
        [4.0, 5.0],
        [7.0, 8.0],
      ]
    ).reshape((1, 2, 2, 1))
    y = pool(x)
    np.testing.assert_allclose(y, expected_y)
    y_grad = jax.grad(lambda x: pool(x).sum())(x)
    expected_grad = jnp.array(
      [
        [0.0, 0.0, 0.0],
        [0.0, 1.0, 1.0],
        [0.0, 1.0, 1.0],
      ]
    ).reshape((1, 3, 3, 1))
    np.testing.assert_allclose(y_grad, expected_grad)

  @parameterized.parameters(
    {'count_include_pad': True}, {'count_include_pad': False}
  )
  def test_avg_pool_padding_same(self, count_include_pad):
    x = jnp.array([1.0, 2.0, 3.0, 4.0]).reshape((1, 2, 2, 1))
    pool = lambda x: nn.avg_pool(
      x, (2, 2), padding='SAME', count_include_pad=count_include_pad
    )
    y = pool(x)
    if count_include_pad:
      expected_y = jnp.array([10.0 / 4, 6.0 / 4, 7.0 / 4, 4.0 / 4]).reshape(
        (1, 2, 2, 1)
      )
    else:
      expected_y = jnp.array([10.0 / 4, 6.0 / 2, 7.0 / 2, 4.0 / 1]).reshape(
        (1, 2, 2, 1)
      )
    np.testing.assert_allclose(y, expected_y)

  def test_pooling_variable_batch_dims(self):
    x = jnp.zeros((1, 8, 32, 32, 3), dtype=jnp.float32)
    y = nn.max_pool(x, (2, 2), (2, 2))

    assert y.shape == (1, 8, 16, 16, 3)

  def test_pooling_no_batch_dims(self):
    x = jnp.zeros((32, 32, 3), dtype=jnp.float32)
    y = nn.max_pool(x, (2, 2), (2, 2))

    assert y.shape == (16, 16, 3)


class NormalizationTest(parameterized.TestCase):
  def test_layer_norm_mask(self):
    key = random.key(0)
    keys = random.split(key)
    x = random.normal(keys[0], (3, 4, 5))
    m = random.choice(keys[1], 2, x.shape).astype(bool)
    m = m.at[..., :2].set(True)  # guarantee at least 2 elements
    x = jnp.where(m, x, jnp.nan)

    module = nn.LayerNorm()
    y, w = module.init_with_output(key, x, mask=m)

    z = y.mean(-1, where=m)
    np.testing.assert_allclose(z, 0, atol=1e-4)

    z = y.var(-1, where=m)
    np.testing.assert_allclose(z, 1, atol=1e-4)

  def test_rms_norm_mask(self):
    key = random.key(0)
    keys = random.split(key)
    x = random.normal(keys[0], (3, 4, 5))
    m = random.choice(keys[1], 2, x.shape).astype(bool)
    m = m.at[..., :1].set(True)  # guarantee at least 1 element
    x = jnp.where(m, x, jnp.nan)

    module = nn.RMSNorm()
    y, w = module.init_with_output(key, x, mask=m)

    z = np.square(y).mean(-1, where=m)
    np.testing.assert_allclose(z, 1, atol=1e-4)

  def test_group_norm_mask(self):
    key = random.key(0)
    keys = random.split(key)
    x = random.normal(keys[0], (13, 3, 5, 7 * 11))
    m = random.choice(keys[1], 2, x.shape).astype(bool)
    m = m.at[..., :2].set(True)  # guarantee at least 2 elements
    x = jnp.where(m, x, jnp.nan)

    module = nn.GroupNorm(7, use_bias=False, use_scale=False)
    y, w = module.init_with_output(key, x, mask=m)

    yr = y.reshape((13, 3, 5, 7, 11))
    mr = m.reshape((13, 3, 5, 7, 11))

    axes = list(range(1, x.ndim - 1)) + [-1]

    z = yr.mean(axes, where=mr)
    np.testing.assert_allclose(z, 0, atol=1e-4)

    z = yr.var(axes, where=mr)
    np.testing.assert_allclose(z, 1, atol=1e-4)

  @parameterized.parameters({'test_mask': True}, {'test_mask': False})
  def test_batch_norm(self, test_mask):
    rng = random.key(0)
    key1, key2, key3 = random.split(rng, 3)
    x = random.normal(key1, (4, 3, 2))
    if test_mask:
      m = random.randint(
        key2, (4, 3, 1), minval=0, maxval=2, dtype=jnp.int32
      ).astype(jnp.bool_)
      x = jnp.where(m, x, jnp.ones_like(x) * jnp.nan)
    else:
      m = None
    model_cls = nn.BatchNorm(momentum=0.9, use_running_average=False)
    y, initial_params = model_cls.init_with_output(key3, x, mask=m)

    mean = y.mean((0, 1), where=m)
    var = y.var((0, 1), where=m)
    np.testing.assert_allclose(mean, np.array([0.0, 0.0]), atol=1e-4)
    np.testing.assert_allclose(var, np.array([1.0, 1.0]), rtol=1e-4)
    _, vars_out = model_cls.apply(
      initial_params, x, mutable=['batch_stats'], mask=m
    )

    ema = vars_out['batch_stats']
    np.testing.assert_allclose(
      ema['mean'], 0.1 * x.mean((0, 1), keepdims=False, where=m), atol=1e-4
    )
    np.testing.assert_allclose(
      ema['var'],
      0.9 + 0.1 * x.var((0, 1), keepdims=False, where=m),
      rtol=1e-4,
    )

  @parameterized.parameters({'test_mask': True}, {'test_mask': False})
  def test_batch_norm_complex(self, test_mask):
    rng = random.key(0)
    key1, key2, key3 = random.split(rng, 3)
    x = random.normal(key1, (4, 3, 2), dtype=jnp.complex64)
    if test_mask:
      m = random.randint(
        key2, (4, 3, 1), minval=0, maxval=2, dtype=jnp.int32
      ).astype(jnp.bool_)
      x = jnp.where(m, x, jnp.ones_like(x) * jnp.nan)
    else:
      m = None
    model_cls = nn.BatchNorm(
      momentum=0.9, use_running_average=False, dtype=jnp.complex64
    )
    y, initial_params = model_cls.init_with_output(key3, x, mask=m)

    mean = y.mean((0, 1), where=m)
    var = y.var((0, 1), where=m)
    np.testing.assert_allclose(mean, np.array([0.0, 0.0]), atol=1e-4)
    np.testing.assert_allclose(var, np.array([1.0, 1.0]), rtol=1e-4)
    self.assertEqual(mean.dtype, jnp.complex64)

    _, vars_out = model_cls.apply(
      initial_params, x, mutable=['batch_stats'], mask=m
    )

    ema = vars_out['batch_stats']
    np.testing.assert_allclose(
      ema['mean'], 0.1 * x.mean((0, 1), keepdims=False, where=m), atol=1e-4
    )
    np.testing.assert_allclose(
      ema['var'],
      0.9 + 0.1 * x.var((0, 1), keepdims=False, where=m),
      rtol=1e-4,
    )

  @parameterized.parameters(
    {'reduction_axes': -1},
    {'reduction_axes': 1},
    {'reduction_axes': (1, 2)},
    {'reduction_axes': (0, 1, 2)},
    {'reduction_axes': -1, 'use_fast_variance': False},
  )
  def test_layer_norm(self, reduction_axes, use_fast_variance=True):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 3, 4))
    if not use_fast_variance:
      x += 1e6  # This blows up fast variance, but should work otherwise.
    model_cls = nn.LayerNorm(
      use_bias=False,
      use_scale=False,
      epsilon=e,
      reduction_axes=reduction_axes,
      use_fast_variance=use_fast_variance,
    )
    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)
    y_one_liner = (
      x - x.mean(axis=reduction_axes, keepdims=True)
    ) * jax.lax.rsqrt(x.var(axis=reduction_axes, keepdims=True) + e)
    np.testing.assert_allclose(y_one_liner, y, atol=1e-4)

  @parameterized.parameters(
    {'reduction_axes': -1}, {'reduction_axes': 1}, {'reduction_axes': (1, 2)}
  )
  def test_rms_norm(self, reduction_axes):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 3, 4))
    model_cls = nn.RMSNorm(
      use_scale=False, epsilon=e, reduction_axes=reduction_axes
    )
    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)
    y_one_liner = x * jax.lax.rsqrt(
      jnp.mean(jax.lax.square(x), axis=reduction_axes, keepdims=True) + e
    )
    np.testing.assert_allclose(y_one_liner, y, atol=1e-4)

  def test_group_norm(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 5, 4, 4, 32))
    model_cls = nn.GroupNorm(
      num_groups=2, use_bias=False, use_scale=False, epsilon=e
    )

    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)

    x_gr = x.reshape([2, 5, 4, 4, 2, 16])
    y_test = (
      x_gr - x_gr.mean(axis=[1, 2, 3, 5], keepdims=True)
    ) * jax.lax.rsqrt(x_gr.var(axis=[1, 2, 3, 5], keepdims=True) + e)
    y_test = y_test.reshape([2, 5, 4, 4, 32])

    np.testing.assert_allclose(y_test, y, atol=1e-4)

  def test_group_norm_unbatched(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 5, 4, 4, 32))
    model_cls = nn.GroupNorm(
        num_groups=2,
        use_bias=False,
        use_scale=False,
        epsilon=e,
        reduction_axes=(0, 1, 3, 4),
    )

    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)

    x_gr = x.reshape([2, 5, 4, 4, 2, 16])
    y_test = (
        x_gr - x_gr.mean(axis=[0, 1, 3, 5], keepdims=True)
    ) * jax.lax.rsqrt(x_gr.var(axis=[0, 1, 3, 5], keepdims=True) + e)
    y_test = y_test.reshape([2, 5, 4, 4, 32])

    np.testing.assert_allclose(y_test, y, atol=1e-4)

  def test_group_norm_batched(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (3, 4, 32))
    model_cls = nn.GroupNorm(
        num_groups=2,
        use_bias=False,
        use_scale=False,
        epsilon=e,
        reduction_axes=(-3, -2, -1),
    )

    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)

    x_stacked = jnp.stack([x] * 5)
    y_stacked = model_cls.apply({}, x_stacked)

    np.testing.assert_allclose(y, y_stacked[0, ...], atol=1e-4)

    x_gr = x_stacked.reshape([5, 3, 4, 2, 16])
    y_test = (x_gr - x_gr.mean(axis=[1, 2, 4], keepdims=True)) * jax.lax.rsqrt(
        x_gr.var(axis=[1, 2, 4], keepdims=True) + e
    )
    y_test = y_test.reshape([5, 3, 4, 32])

    np.testing.assert_allclose(y_test, y_stacked, atol=1e-4)

  def test_group_norm_raises(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 5, 4, 4, 32))
    model_cls = nn.GroupNorm(
        num_groups=3, use_bias=False, use_scale=False, epsilon=e
    )

    with self.assertRaises(ValueError):
      model_cls.init_with_output(key2, x)

  def test_group_norm_raises_incorrect_reduction_axes(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 5, 4, 4, 32))
    model_cls = nn.GroupNorm(
        num_groups=3,
        use_bias=False,
        use_scale=False,
        epsilon=e,
        reduction_axes=(0, 1, 2, 3),
    )

    with self.assertRaises(ValueError):
      model_cls.init_with_output(key2, x)

  def test_batch_norm_multi_init(self):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        norm = nn.BatchNorm(
          name='norm',
          use_running_average=False,
          axis_name='batch',
        )
        x = norm(x)
        return x, norm(x)

    key = random.key(0)
    model = Foo()
    x = random.normal(random.key(1), (2, 4))
    (y1, y2), _ = model.init_with_output(key, x)
    np.testing.assert_allclose(y1, y2, rtol=1e-4)

  @parameterized.parameters(
    {'feature_axes': -1},
    {'feature_axes': (1, 2)},
    {'feature_axes': (1, 2, 3)},
    {'feature_axes': -1, 'use_fast_variance': False},
  )
  def test_instance_norm(self, feature_axes, use_fast_variance=True):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    e = 1e-5
    x = random.normal(key1, (2, 3, 4, 5))
    if not use_fast_variance:
      x += 1e4  # This blows up fast variance, but should work otherwise.
    model_cls = nn.InstanceNorm(
      use_bias=False,
      use_scale=False,
      epsilon=e,
      feature_axes=feature_axes,
      use_fast_variance=use_fast_variance,
    )
    y, _ = model_cls.init_with_output(key2, x)
    self.assertEqual(x.dtype, y.dtype)
    self.assertEqual(x.shape, y.shape)

    canonicalized_feature_axes = [
      i if i >= 0 else (x.ndim + i)
      for i in (
        feature_axes if isinstance(feature_axes, tuple) else (feature_axes,)
      )
    ]
    reduction_axes = [
      i for i in range(1, x.ndim) if i not in canonicalized_feature_axes
    ]
    y_one_liner = (
      x - x.mean(axis=reduction_axes, keepdims=True)
    ) * jax.lax.rsqrt(x.var(axis=reduction_axes, keepdims=True) + e)

    np.testing.assert_allclose(y_one_liner, y, atol=1e-6)

  @parameterized.parameters(
    {'feature_axes': 0},
    {'feature_axes': -4},
    {'feature_axes': (0, 3)},
    {'feature_axes': (2, -4)},
  )
  def test_instance_norm_raise_error(self, feature_axes):
    with self.assertRaisesRegex(
      ValueError,
      'The channel axes cannot include the leading dimension '
      'as this is assumed to be the batch axis.',
    ):
      x = jax.random.normal(jax.random.key(0), (2, 3, 4, 5))
      layer = nn.InstanceNorm(feature_axes=feature_axes)
      _ = layer.init(jax.random.key(1), x)

  @parameterized.parameters(
    {
      'layer1': nn.LayerNorm(feature_axes=(1, 2)),
      'layer2': nn.InstanceNorm(feature_axes=(1, 2)),
    },
    {
      'layer1': nn.LayerNorm(reduction_axes=(1, 2), feature_axes=-1),
      'layer2': nn.InstanceNorm(feature_axes=-1),
    },
    {
      'layer1': nn.LayerNorm(
        reduction_axes=-2,
        feature_axes=(1, 3),
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform(),
      ),
      'layer2': nn.InstanceNorm(
        feature_axes=(1, -1),
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform(),
      ),
    },
    {
      'layer1': nn.LayerNorm(
        reduction_axes=(1, 2, 3),
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform(),
      ),
      'layer2': nn.GroupNorm(
        num_groups=1,
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform()
      ),
    },
    {
      'layer1': nn.InstanceNorm(
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform(),
      ),
      'layer2': nn.GroupNorm(
        num_groups=None,
        group_size=1,
        bias_init=nn.initializers.uniform(),
        scale_init=nn.initializers.uniform(),
      ),
    },
  )
  def test_normalization_equivalence(self, layer1, layer2):
    x = jax.random.normal(jax.random.key(0), (2, 3, 4, 5))
    layer1_variables = layer1.init(jax.random.key(1), x)
    layer2_variables = layer2.init(jax.random.key(1), x)
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(
                lambda v1, v2: (v1 == v2).all(),
                layer1_variables,
                layer2_variables,
            )
        )
    )

    layer1_y = layer1.apply(layer1_variables, x)
    layer2_y = layer2.apply(layer2_variables, x)
    np.testing.assert_allclose(layer1_y, layer2_y, atol=1e-7)

  @parameterized.parameters(
    {
      'model_index': 0,
      'key_paths': {'Dense_1/kernel/u', 'Dense_1/kernel/sigma'},
    },
    {
      'model_index': 1,
      'key_paths': {'Conv_0/kernel/u', 'Conv_0/kernel/sigma'},
    },
    {
      'model_index': 2,
      'key_paths': {
        'MultiHeadDotProductAttention_0/key/bias/u',
        'MultiHeadDotProductAttention_0/key/kernel/u',
        'MultiHeadDotProductAttention_0/out/kernel/u',
        'MultiHeadDotProductAttention_0/query/bias/u',
        'MultiHeadDotProductAttention_0/query/kernel/u',
        'MultiHeadDotProductAttention_0/value/bias/u',
        'MultiHeadDotProductAttention_0/value/kernel/u',
        'MultiHeadDotProductAttention_0/key/bias/sigma',
        'MultiHeadDotProductAttention_0/key/kernel/sigma',
        'MultiHeadDotProductAttention_0/out/kernel/sigma',
        'MultiHeadDotProductAttention_0/query/bias/sigma',
        'MultiHeadDotProductAttention_0/query/kernel/sigma',
        'MultiHeadDotProductAttention_0/value/bias/sigma',
        'MultiHeadDotProductAttention_0/value/kernel/sigma',
      },
    },
  )
  def test_spectral_norm_train(self, model_index, key_paths):
    class FooDense(nn.Module):
      @nn.compact
      def __call__(self, x, train):
        x = nn.Dense(8)(x)
        x = nn.SpectralNorm(nn.Dense(6))(x, update_stats=train)
        x = nn.Dense(4)(x)
        return x

    class FooConv(nn.Module):
      @nn.compact
      def __call__(self, x, train):
        x = nn.Dense(9)(x)
        x = x.reshape((1, 3, 3))
        x = nn.SpectralNorm(nn.Conv(2, kernel_size=(2, 2)))(
          x, update_stats=train
        )
        x = x.reshape(1, -1)
        x = nn.Dense(4)(x)
        return x

    class FooAttention(nn.Module):
      @nn.compact
      def __call__(self, x, train):
        a = nn.Dense(4)(x)
        b = nn.Dense(4)(x)
        x = nn.SpectralNorm(nn.attention.MultiHeadDotProductAttention(4))(
          a, b, update_stats=train
        )
        x = nn.Dense(4)(x)
        return x

    key1, key2, key3 = random.split(random.PRNGKey(0), 3)
    x = random.normal(key1, (1, 4))
    y = random.normal(key2, (1, 4))

    model_cls = (FooDense, FooConv, FooAttention)[model_index]
    variables = model_cls().init(key3, x, train=False)
    params, batch_stats = variables['params'], variables['batch_stats']
    self.assertEqual(key_paths, batch_stats['SpectralNorm_0'].keys())

    class TrainState(train_state.TrainState):
      batch_stats: Any

    state = TrainState.create(
      apply_fn=model_cls().apply,
      params=params,
      batch_stats=batch_stats,
      tx=optax.adam(1e-3),
    )

    @jax.jit
    def train_step(state, batch):
      def loss_fn(params):
        logits, updates = state.apply_fn(
          {'params': params, 'batch_stats': state.batch_stats},
          x=batch['image'],
          train=True,
          mutable=['batch_stats'],
        )
        loss = jnp.mean(
          optax.l2_loss(predictions=logits, targets=batch['label'])
        )
        return loss, updates

      grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
      (loss, updates), grads = grad_fn(state.params)
      state = state.apply_gradients(grads=grads)
      state = state.replace(batch_stats=updates['batch_stats'])
      return state, loss

    prev_loss = float('inf')
    for _ in range(10):
      state, loss = train_step(state, {'image': x, 'label': y})
      self.assertLess(loss, prev_loss)
      prev_loss = loss

  @parameterized.parameters(
    {'n_steps': 1, 'update_stats': True, 'result': 4.0},
    {'n_steps': 3, 'update_stats': True, 'result': 4.0},
    {'n_steps': 10, 'update_stats': True, 'result': 4.0},
    {'n_steps': 1, 'update_stats': False, 'result': 1.0},
  )
  def test_spectral_norm_sigma(self, n_steps, update_stats, result):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x, train):
        x = nn.SpectralNorm(nn.Dense(8, use_bias=False), n_steps=n_steps)(
          x, update_stats=train
        )
        return x

    x = jnp.ones((1, 8))
    model_cls = Foo()
    variables = model_cls.init(random.PRNGKey(0), x, train=False)
    params, batch_stats = variables['params'], variables['batch_stats']
    params = jax.tree_util.tree_map(lambda x: 4 * jnp.eye(*x.shape), params)
    _, updates = model_cls.apply(
      {'params': params, 'batch_stats': batch_stats},
      x=x,
      train=update_stats,
      mutable=True,
    )
    np.testing.assert_allclose(
      updates['batch_stats']['SpectralNorm_0']['Dense_0/kernel/sigma'],
      result,
      atol=1e-3,
    )

  @parameterized.parameters(
    {'error_on_non_matrix': True}, {'error_on_non_matrix': False}
  )
  def test_spectral_norm_3d_tensor(self, error_on_non_matrix):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x, train):
        x = nn.SpectralNorm(
          nn.DenseGeneral((3, 4), use_bias=False),
          error_on_non_matrix=error_on_non_matrix,
        )(x, update_stats=train)
        return x

    x = jnp.ones((1, 2))
    model_cls = Foo()

    if error_on_non_matrix:
      with self.assertRaisesRegex(
        ValueError, 'Input is 3D but error_on_non_matrix is True'
      ):
        _ = model_cls.init(random.PRNGKey(0), x, train=False)
    else:
      _ = model_cls.init(random.PRNGKey(0), x, train=False)

  @parameterized.parameters(
    {'feature_axes': -1, 'reduction_axes': 0, 'variable_filter': {'kernel'}},
    {'feature_axes': 0, 'reduction_axes': 1, 'variable_filter': {'kernel'}},
    {
      'feature_axes': (0, 1),
      'reduction_axes': (),
      'variable_filter': {'kernel'},
    },
    {
      'feature_axes': (),
      'reduction_axes': (0, 1),
      'variable_filter': {'kernel'},
    },
    {
      'feature_axes': None,
      'reduction_axes': (0, 1),
      'variable_filter': {'kernel'},
    },
    {'feature_axes': 0, 'reduction_axes': (), 'variable_filter': {'bias'}},
    {'feature_axes': (), 'reduction_axes': -1, 'variable_filter': {'bias'}},
  )
  def test_manual_weight_norm(
    self, feature_axes, reduction_axes, variable_filter
  ):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        return nn.WeightNorm(
          nn.Dense(2, bias_init=nn.initializers.normal()),
          feature_axes=feature_axes,
          variable_filter=variable_filter,
        )(x)

    key1, key2 = jax.random.split(jax.random.key(1))
    x = jax.random.normal(key1, (1, 3))
    module = Foo()
    v = module.init(key2, x)
    v = jax.tree_util.tree_map(lambda x: x + 0.5, v)
    out = module.apply(v, x)

    kernel = v['params']['Dense_0']['kernel']
    if 'kernel' in variable_filter:
      kernel /= jnp.sqrt(jnp.sum(kernel**2, axis=reduction_axes, keepdims=True))
      kernel_scale = jnp.expand_dims(
        v['params']['WeightNorm_0']['Dense_0/kernel/scale'],
        axis=reduction_axes,
      )
    else:
      kernel_scale = 1
    bias = v['params']['Dense_0']['bias']
    if 'bias' in variable_filter:
      bias /= jnp.sqrt(jnp.sum(bias**2, axis=reduction_axes, keepdims=True))
      bias_scale = jnp.expand_dims(
        v['params']['WeightNorm_0']['Dense_0/bias/scale'], axis=reduction_axes
      )
    else:
      bias_scale = 1
    manual_out = jnp.dot(x, kernel_scale * kernel) + (
      bias_scale * bias
    ).reshape(1, -1)

    self.assertTrue(jnp.allclose(out, manual_out))

  @parameterized.parameters(
    {
      'variable_filters': ({}, None, {'kernel', 'bias'}, {'Bar'}),
      'key_paths': {
        'Bar_0/Baz_0/Dense_0/kernel/scale',
        'Bar_0/Baz_0/Dense_0/bias/scale',
        'Bar_0/Dense_0/kernel/scale',
        'Bar_0/Dense_0/bias/scale',
        'Bar_0/Baz_1/Dense_0/kernel/scale',
        'Bar_0/Baz_1/Dense_0/bias/scale',
        'Bar_0/Dense_1/kernel/scale',
        'Bar_0/Dense_1/bias/scale',
      },
    },
    {
      'variable_filters': ({'kernel'},),
      'key_paths': {
        'Bar_0/Baz_0/Dense_0/kernel/scale',
        'Bar_0/Dense_0/kernel/scale',
        'Bar_0/Baz_1/Dense_0/kernel/scale',
        'Bar_0/Dense_1/kernel/scale',
      },
    },
    {
      'variable_filters': ({'Baz', 'kernel'},),
      'key_paths': {
        'Bar_0/Baz_0/Dense_0/kernel/scale',
        'Bar_0/Baz_0/Dense_0/bias/scale',
        'Bar_0/Dense_0/kernel/scale',
        'Bar_0/Baz_1/Dense_0/kernel/scale',
        'Bar_0/Baz_1/Dense_0/bias/scale',
        'Bar_0/Dense_1/kernel/scale',
      },
    },
  )
  def test_weight_norm_variable_filter(self, variable_filters, key_paths):
    class Baz(nn.Module):
      @nn.compact
      def __call__(self, x):
        return nn.Dense(2)(x)

    class Bar(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = Baz()(x)
        x = nn.Dense(3)(x)
        x = Baz()(x)
        x = nn.Dense(3)(x)
        return x

    for variable_filter in variable_filters:

      class Foo(nn.Module):
        @nn.compact
        def __call__(self, x):
          return nn.WeightNorm(Bar(), variable_filter=variable_filter)(x)

      v = Foo().init(jax.random.key(0), jnp.ones((1, 4)))
      self.assertEqual(key_paths, v['params']['WeightNorm_0'].keys())

  @parameterized.parameters(
    {'model_index': 0, 'key_paths': {'Dense_1/kernel/scale'}},
    {'model_index': 1, 'key_paths': {'Conv_0/kernel/scale'}},
    {
      'model_index': 2,
      'key_paths': {
        'MultiHeadDotProductAttention_0/key/kernel/scale',
        'MultiHeadDotProductAttention_0/out/kernel/scale',
        'MultiHeadDotProductAttention_0/query/kernel/scale',
        'MultiHeadDotProductAttention_0/value/kernel/scale',
      },
    },
  )
  def test_weight_norm_train(self, model_index, key_paths):
    class FooDense(nn.Module):
      @nn.compact
      def __call__(
        self,
        x,
      ):
        x = nn.Dense(8)(x)
        x = nn.WeightNorm(nn.Dense(6))(x)
        x = nn.Dense(4)(x)
        return x

    class FooConv(nn.Module):
      @nn.compact
      def __call__(
        self,
        x,
      ):
        x = nn.Dense(9)(x)
        x = x.reshape((1, 3, 3))
        x = nn.WeightNorm(nn.Conv(2, kernel_size=(2, 2)))(x)
        x = x.reshape(1, -1)
        x = nn.Dense(4)(x)
        return x

    class FooAttention(nn.Module):
      @nn.compact
      def __call__(self, x):
        a = nn.Dense(4)(x)
        b = nn.Dense(4)(x)
        x = nn.WeightNorm(nn.attention.MultiHeadDotProductAttention(4))(a, b)
        x = nn.Dense(4)(x)
        return x

    key1, key2, key3 = random.split(random.PRNGKey(0), 3)
    x = random.normal(key1, (1, 4))
    y = random.normal(key2, (1, 4))

    model_cls = (FooDense, FooConv, FooAttention)[model_index]
    params = model_cls().init(key3, x)['params']
    self.assertEqual(key_paths, params['WeightNorm_0'].keys())

    state = train_state.TrainState.create(
      apply_fn=model_cls().apply,
      params=params,
      tx=optax.adam(1e-3),
    )

    @jax.jit
    def train_step(state, batch):
      def loss_fn(params):
        logits = state.apply_fn(
          {'params': params},
          x=batch['image'],
        )
        loss = jnp.mean(
          optax.l2_loss(predictions=logits, targets=batch['label'])
        )
        return loss

      grad_fn = jax.value_and_grad(loss_fn)
      loss, grads = grad_fn(state.params)
      state = state.apply_gradients(grads=grads)
      return state, loss

    prev_loss = float('inf')
    for _ in range(10):
      state, loss = train_step(state, {'image': x, 'label': y})
      self.assertLess(loss, prev_loss)
      prev_loss = loss


class StochasticTest(parameterized.TestCase):
  def test_dropout(self):
    rng = random.key(0)
    key1, key2 = random.split(rng)
    module = nn.Dropout(rate=0.5)
    y1 = module.apply(
      {}, jnp.ones((20, 20)), deterministic=False, rngs={'dropout': key1}
    )
    y2 = module.apply(
      {}, jnp.ones((20, 20)), deterministic=False, rngs={'dropout': key2}
    )
    self.assertFalse(np.all(y1 == y2))

    y1 = module.apply(
      {}, jnp.ones((20, 20)), deterministic=True, rngs={'dropout': key1}
    )
    y2 = module.apply(
      {}, jnp.ones((20, 20)), deterministic=True, rngs={'dropout': key2}
    )
    self.assertTrue(np.all(y1 == y2))

  def test_dropout_rate_stats(self):
    rootkey = random.key(0)
    for rate in np.arange(0.1, 1.0, 0.1):
      rootkey, subkey = random.split(rootkey)
      module = nn.Dropout(rate=rate)
      n_trials = 10
      nonzero_counts = 0
      for key in random.split(subkey, n_trials):
        y = module.apply(
          {}, jnp.ones((100, 100)), deterministic=False, rngs={'dropout': key}
        )
        nonzero_counts += np.sum(y > 0.0)
      all_counts = np.prod((100, 100, n_trials))
      frac = np.sum(nonzero_counts) / all_counts
      keep_rate = 1.0 - rate
      # just check within 4 sigma.
      delta = 4 * np.sqrt(rate * keep_rate) / np.sqrt(all_counts)
      self.assertTrue(keep_rate - delta < frac < keep_rate + delta)

  def test_dropout_rate_limits(self):
    rng = random.key(0)
    key1, key2, key3 = random.split(rng, 3)
    inputs = jnp.ones((20, 20))
    d0 = nn.Dropout(rate=0.0)
    y1 = d0.apply({}, inputs, deterministic=False, rngs={'dropout': key1})
    np.testing.assert_array_equal(y1, inputs)
    d1 = nn.Dropout(rate=1.0)
    y2 = d1.apply({}, inputs, deterministic=False, rngs={'dropout': key2})
    np.testing.assert_array_equal(y2, np.zeros_like(inputs))
    # ensure gradient of rate==1.0 case is non-NaN
    fn = lambda x, k: d1.apply({}, x, rngs={'dropout': k}, deterministic=False)
    res = jax.grad(lambda x, k: jnp.sum(fn(x, k)))(inputs, key3)
    self.assertFalse(np.isnan(res).any())

  @parameterized.parameters(
    {
      'num_dims': 2,
      'broadcast_dims': (1,),
      'slice_fn': lambda out, i: out[i, :],
      'summed_total': 2 * 10,
    },
    {
      'num_dims': 2,
      'broadcast_dims': (0,),
      'slice_fn': lambda out, i: out[:, i],
      'summed_total': 2 * 10,
    },
    {
      'num_dims': 3,
      'broadcast_dims': (1, 2),
      'slice_fn': lambda out, i: out[i, :, :],
      'summed_total': 2 * 10 * 10,
    },
    {
      'num_dims': 3,
      'broadcast_dims': (1,),
      'slice_fn': lambda out, i, j: out[i, :, j],
      'summed_total': 2 * 10,
    },
    {
      'num_dims': 4,
      'broadcast_dims': (0, 2, 3),
      'slice_fn': lambda out, i: out[:, i, :, :],
      'summed_total': 2 * 10 * 10 * 10,
    },
    {
      'num_dims': 4,
      'broadcast_dims': (0, 1),
      'slice_fn': lambda out, i, j: out[:, :, i, j],
      'summed_total': 2 * 10 * 10,
    },
    {
      'num_dims': 4,
      'broadcast_dims': (3,),
      'slice_fn': lambda out, i, j, k: out[i, j, k, :],
      'summed_total': 2 * 10,
    },
  )
  def test_dropout_broadcast(
    self, num_dims, broadcast_dims, slice_fn, summed_total
  ):
    module = nn.Dropout(
      rate=0.5, broadcast_dims=broadcast_dims, deterministic=False
    )
    x = jnp.ones((10,) * num_dims)
    out = module.apply({}, x, rngs={'dropout': random.key(0)})

    for i in range(10):
      if num_dims - len(broadcast_dims) >= 2:
        for j in range(10):
          if num_dims - len(broadcast_dims) >= 3:
            for k in range(10):
              self.assertTrue(slice_fn(out, i, j, k).sum() in (0, summed_total))
          else:
            self.assertTrue(slice_fn(out, i, j).sum() in (0, summed_total))
      else:
        self.assertTrue(slice_fn(out, i).sum() in (0, summed_total))

  def test_dropout_manual_rng(self):
    def clone(key):
      if hasattr(jax.random, 'clone'):
        # JAX v0.4.26+
        return jax.random.clone(key)
      return key
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        key = self.make_rng('dropout')
        x1 = nn.Dropout(rate=0.5, deterministic=False)(x, rng=key)
        x2 = nn.Dropout(rate=0.5, deterministic=False)(x, rng=clone(key))
        return x1, x2

    module = Foo()
    x1, x2 = module.apply(
      {}, jnp.ones((20, 20)), rngs={'dropout': random.key(0)}
    )

    np.testing.assert_array_equal(x1, x2)


# TODO(flax-dev): add integration tests for RNN cells
class RecurrentTest(parameterized.TestCase):
  def test_lstm(self):
    lstm = nn.LSTMCell(features=4)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3))
    c0, h0 = lstm.initialize_carry(rng, x.shape)
    self.assertEqual(c0.shape, (2, 4))
    self.assertEqual(h0.shape, (2, 4))
    (carry, y), initial_params = lstm.init_with_output(key2, (c0, h0), x)
    self.assertEqual(carry[0].shape, (2, 4))
    self.assertEqual(carry[1].shape, (2, 4))
    np.testing.assert_allclose(y, carry[1])
    param_shapes = jax.tree_util.tree_map(np.shape, initial_params['params'])
    self.assertEqual(
      param_shapes,
      {
        'ii': {'kernel': (3, 4)},
        'if': {'kernel': (3, 4)},
        'ig': {'kernel': (3, 4)},
        'io': {'kernel': (3, 4)},
        'hi': {'kernel': (4, 4), 'bias': (4,)},
        'hf': {'kernel': (4, 4), 'bias': (4,)},
        'hg': {'kernel': (4, 4), 'bias': (4,)},
        'ho': {'kernel': (4, 4), 'bias': (4,)},
      },
    )

  @parameterized.parameters(
    {
      'module_cls': nn.SimpleCell,
      'expected_param_shapes': {
        'i': {'kernel': (3, 4), 'bias': (4,)},
        'h': {'kernel': (4, 4)},
      },
    },
    {
      'module_cls': nn.GRUCell,
      'expected_param_shapes': {
        'ir': {'kernel': (3, 4), 'bias': (4,)},
        'iz': {'kernel': (3, 4), 'bias': (4,)},
        'in': {'kernel': (3, 4), 'bias': (4,)},
        'hr': {'kernel': (4, 4)},
        'hz': {'kernel': (4, 4)},
        'hn': {'kernel': (4, 4), 'bias': (4,)},
      },
    },
    {
      'module_cls': nn.MGUCell,
      'expected_param_shapes': {
        'if': {'kernel': (3, 4), 'bias': (4,)},
        'in': {'kernel': (3, 4), 'bias': (4,)},
        'hf': {'kernel': (4, 4)},
        'hn': {'kernel': (4, 4), 'bias': (4,)},
      },
    },
  )
  def test_gated_units(self, module_cls, expected_param_shapes):
    module = module_cls(features=4)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3))
    carry0 = module.initialize_carry(rng, x.shape)
    self.assertEqual(carry0.shape, (2, 4))
    (carry, y), initial_params = module.init_with_output(key2, carry0, x)
    self.assertEqual(carry.shape, (2, 4))
    np.testing.assert_allclose(y, carry)
    param_shapes = jax.tree_util.tree_map(np.shape, initial_params['params'])
    self.assertEqual(
      param_shapes,
      expected_param_shapes,
    )
    if module_cls == nn.MGUCell:
      self.assertTrue(
        (initial_params['params']['if']['bias'] == jnp.ones((4,))).all()
      )
      self.assertTrue(
        (initial_params['params']['in']['bias'] == jnp.zeros((4,))).all()
      )
      self.assertTrue(
        (initial_params['params']['hn']['bias'] == jnp.zeros((4,))).all()
      )

  @parameterized.parameters(
    {'module_cls': nn.SimpleCell},
    {'module_cls': nn.GRUCell},
    {'module_cls': nn.MGUCell},
  )
  def test_complex_input_gated_units(self, module_cls):
    module_instance = module_cls(features=4)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3), dtype=jnp.complex64)
    carry0 = module_instance.initialize_carry(rng, x.shape)
    self.assertEqual(carry0.shape, (2, 4))
    (carry, y), _ = module_instance.init_with_output(key2, carry0, x)
    self.assertEqual(carry.dtype, jnp.complex64)
    self.assertEqual(y.dtype, jnp.complex64)

  def test_convlstm(self):
    lstm = nn.ConvLSTMCell(features=6, kernel_size=(3, 3))
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 4, 4, 3))
    c0, h0 = lstm.initialize_carry(rng, x.shape)
    self.assertEqual(c0.shape, (2, 4, 4, 6))
    self.assertEqual(h0.shape, (2, 4, 4, 6))
    (carry, y), initial_params = lstm.init_with_output(key2, (c0, h0), x)
    self.assertEqual(carry[0].shape, (2, 4, 4, 6))
    self.assertEqual(carry[1].shape, (2, 4, 4, 6))
    np.testing.assert_allclose(y, carry[1])
    param_shapes = jax.tree_util.tree_map(np.shape, initial_params['params'])
    self.assertEqual(
      param_shapes,
      {
        'hh': {'bias': (6 * 4,), 'kernel': (3, 3, 6, 6 * 4)},
        'ih': {'bias': (6 * 4,), 'kernel': (3, 3, 3, 6 * 4)},
      },
    )

  def test_optimized_lstm_cell_matches_regular(self):
    # Create regular LSTMCell.
    lstm = nn.LSTMCell(features=4)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3))
    c0, h0 = lstm.initialize_carry(rng, x.shape)
    self.assertEqual(c0.shape, (2, 4))
    self.assertEqual(h0.shape, (2, 4))
    (_, y), lstm_params = lstm.init_with_output(key2, (c0, h0), x)

    # Create OptimizedLSTMCell.
    lstm_opt = nn.OptimizedLSTMCell(features=4)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3))
    c0, h0 = lstm_opt.initialize_carry(rng, x.shape)
    self.assertEqual(c0.shape, (2, 4))
    self.assertEqual(h0.shape, (2, 4))
    (_, y_opt), lstm_opt_params = lstm_opt.init_with_output(key2, (c0, h0), x)

    np.testing.assert_allclose(y, y_opt, rtol=1e-6)
    check_eq(lstm_params, lstm_opt_params)

  def test_mgu_reset_gate(self):
    module = nn.MGUCell(features=4, reset_gate=False)
    rng = random.key(0)
    rng, key1, key2 = random.split(rng, 3)
    x = random.normal(key1, (2, 3))
    carry0 = module.initialize_carry(rng, x.shape)
    (carry, y), v = module.init_with_output(key2, carry0, x)

    self.assertIn('kernel', v['params']['hn'])
    self.assertNotIn('bias', v['params']['hn'])

    f = jax.nn.sigmoid(
      jnp.dot(x, v['params']['if']['kernel'])
      + v['params']['if']['bias'].reshape(1, -1)
      + jnp.dot(carry0, v['params']['hf']['kernel'])
    )
    n = jax.nn.tanh(
      jnp.dot(x, v['params']['in']['kernel'])
      + v['params']['in']['bias'].reshape(1, -1)
      + jnp.dot(carry0, v['params']['hn']['kernel'])
    )
    expected_out = (1 - f) * n + f * carry0
    np.testing.assert_allclose(y, expected_out)


class IdsTest(absltest.TestCase):
  def test_hashable(self):
    id1 = ids.uuid()
    id2 = ids.uuid()
    self.assertEqual(id1, id1)
    self.assertNotEqual(id1, id2)
    self.assertNotEqual(hash(id1), hash(id2))
    id1c = copy.copy(id1)
    id1dc = copy.deepcopy(id1)
    self.assertNotEqual(hash(id1), hash(id1c))
    self.assertNotEqual(hash(id1), hash(id1dc))


class Fp8Test(parameterized.TestCase):
  def test_fp8_dot_general_injection(self):
    # Used to cast the inputs to be representable in FP8, so that the difference
    # of the results from the original gemm and fp8 gemm is small.
    cast_to_representable = functools.partial(
      fp8_ops.quantize_dequantize,
      scale=jnp.ones((1,)),
      compute_dtype=jnp.float32,
    )

    init_key, random_key = random.split(random.PRNGKey(seed=123), 2)
    x = cast_to_representable(
      random.uniform(random_key, (16, 32)), jnp.float8_e4m3fn
    )
    dy = cast_to_representable(
      random.uniform(random_key, (16, 64)), jnp.float8_e5m2
    )

    def run(fp8_injection, expected_shapes):
      p = nn.DenseGeneral(features=64, name='dense')
      if fp8_injection:
        p.dot_general_cls = nn.Fp8DotGeneralOp

      init_fn = jax.jit(p.init_with_output)
      y, initial_vars = init_fn(init_key, x)
      var_shapes = jax.tree_util.tree_map(jnp.shape, initial_vars)
      self.assertEqual(var_shapes, expected_shapes)

      def _train(variables, x):
        y = p.apply(variables, x)
        loss = y * dy
        return jnp.mean(loss)

      train_fn = jax.jit(jax.value_and_grad(_train, argnums=[0, 1]))
      outputs, grads = train_fn(initial_vars, x)
      return outputs, grads

    expected_shapes_original = {
      'params': {'kernel': (32, 64), 'bias': (64,)},
    }
    expected_shapes_new = {
      'params': {'kernel': (32, 64), 'bias': (64,)},
      fp8_ops.OVERWRITE_WITH_GRADIENT: {
        'Fp8DotGeneralOp_0': {
          'input_amax_history': (1024,),
          'kernel_amax_history': (1024,),
          'output_grad_amax_history': (1024,),
          'input_scale': (1,),
          'kernel_scale': (1,),
          'output_grad_scale': (1,),
        }
      },
    }

    output1a, output1b = run(False, expected_shapes_original)
    output2a, output2b = run(True, expected_shapes_new)
    dw1, dw2 = output1b[0]['params']['kernel'], output2b[0]['params']['kernel']
    dx1, dx2 = output1b[1], output2b[1]

    np.testing.assert_allclose(output1a, output2a, atol=1e-02)
    np.testing.assert_allclose(dw1, dw2, atol=1e-04)
    np.testing.assert_allclose(dx1, dx2, atol=1e-04)

  def test_fp8_train_state(self):
    key, init_key, random_key = random.split(random.PRNGKey(seed=123), 3)
    x = random.uniform(random_key, (16, 16), dtype=jnp.float32)
    dense = nn.DenseGeneral(
      features=32, use_bias=True, dot_general_cls=nn.Fp8DotGeneralOp
    )

    init_fn = jax.jit(dense.init)
    variables = init_fn(init_key, x)
    opt = optax.adam(learning_rate=0.1)
    state = train_state.TrainState.create(
      params=variables, tx=opt, apply_fn=dense.apply
    )

    def _roll_and_update(amax_h, update):
      return jnp.roll(amax_h, shift=-1, axis=0).at[0].set(update)

    def _train_loss(state, x, dy):
      def loss_fn(vars):
        y = state.apply_fn(vars, x)
        loss = y * dy.astype(y.dtype)
        return jnp.sum(loss)

      grad_fn = jax.grad(loss_fn)
      grads = grad_fn(state.params)
      state = state.apply_gradients(grads=grads)
      return state

    train_fn = jax.jit(_train_loss)

    scale_x, amax_history_x = jnp.ones(()), jnp.zeros((1024,))
    scale_k, amax_history_k = jnp.ones(()), jnp.zeros((1024,))
    scale_g, amax_history_g = jnp.ones(()), jnp.zeros((1024,))
    e4m3_max = jnp.finfo(jnp.float8_e4m3fn).max.astype(jnp.float32)
    e5m2_max = jnp.finfo(jnp.float8_e5m2).max.astype(jnp.float32)

    for _ in range(5):
      key, random_key = random.split(key, 2)
      x = random.normal(random_key, (16, 16), dtype=jnp.float32)
      g = random.normal(random_key, (16, 32), dtype=jnp.float32)
      k = state.params['params']['kernel']

      # Manually compute the expected amax history and scaling factors.
      amax_from_history_x = jnp.max(amax_history_x, axis=0)
      amax_from_history_k = jnp.max(amax_history_k, axis=0)
      amax_from_history_g = jnp.max(amax_history_g, axis=0)
      scale_x = fp8_ops.compute_scale(amax_from_history_x, scale_x, e4m3_max)
      scale_k = fp8_ops.compute_scale(amax_from_history_k, scale_k, e4m3_max)
      scale_g = fp8_ops.compute_scale(amax_from_history_g, scale_g, e5m2_max)
      amax_history_x = _roll_and_update(amax_history_x, jnp.max(jnp.abs(x)))
      amax_history_k = _roll_and_update(amax_history_k, jnp.max(jnp.abs(k)))
      amax_history_g = _roll_and_update(amax_history_g, jnp.max(jnp.abs(g)))

      state = train_fn(state, x, g)

      rtol, atol = 0.001, 0.001
      fp8_vars = state.params[fp8_ops.OVERWRITE_WITH_GRADIENT][
        'Fp8DotGeneralOp_0'
      ]
      np.testing.assert_allclose(
        fp8_vars['input_amax_history'],
        amax_history_x,
        rtol=rtol,
        atol=atol,
      )
      np.testing.assert_allclose(
        fp8_vars['kernel_amax_history'],
        amax_history_k,
        rtol=rtol,
        atol=atol,
      )
      np.testing.assert_allclose(
        fp8_vars['output_grad_amax_history'],
        amax_history_g,
        rtol=rtol,
        atol=atol,
      )

      np.testing.assert_allclose(fp8_vars['input_scale'][0], scale_x)
      np.testing.assert_allclose(fp8_vars['kernel_scale'][0], scale_k)
      np.testing.assert_allclose(fp8_vars['output_grad_scale'][0], scale_g)

  @parameterized.parameters([True, False])
  def test_fp8_meta_dtype(self, use_jit):
    if not use_jit and not fp8_ops.CAN_USE_EARRAY:
      self.skipTest("TODO: requires newer jax that has earray")
    f32 = jnp.dtype('float32')
    fm32 = fp8_ops.fm32

    # Create a scan loop with reused ah_f32 and sf_f32. So, the autograd will
    # accumulate the grads of them. We expect the max op (rather than add op)
    # for the accumulation by converting them to fm32 dtype.
    def outer(x, ah_f32, sf_f32):
      ah_fm32 = jax.lax.convert_element_type(ah_f32, fm32)
      sf_fm32 = jax.lax.convert_element_type(sf_f32, fm32)
      array_x = jnp.array([x], f32)
      def body_fun(carry, _):
        carry = fp8_ops.in_qdq(f32, carry, sf_fm32, ah_fm32)
        return carry, None
      array_x, _ = jax.lax.scan(body_fun, array_x, None, length=3)
      return array_x[0]

    outer_fn = jax.grad(outer, (0, 1, 2))
    if use_jit:
      outer_fn = jax.jit(outer_fn)
    ah = jnp.array([0., 0., 0.], f32)
    sf = jnp.array([1.], f32)
    # 1st iteration
    grads, new_ah, new_sf = outer_fn(2.0, ah, sf)
    np.testing.assert_allclose(new_ah, [2., 0., 0.])
    np.testing.assert_allclose(new_sf, [1.])
    # 2nd iteration
    grads, new_ah, new_sf = outer_fn(3., new_ah, new_sf)
    np.testing.assert_allclose(new_ah, [3., 0., 2.])
    np.testing.assert_allclose(new_sf, [2. / 448])
    # 3rd iteration
    grads, new_ah, new_sf = outer_fn(4., new_ah, new_sf)
    np.testing.assert_allclose(new_ah, [4., 2., 3.])
    np.testing.assert_allclose(new_sf, [3. / 448])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

import jax
import jax.numpy as jnp
from jax import lax

from flax import nnx
from flax.nnx.nnx import rnglib
from flax.nnx.nnx.module import Module, first_from
from flax.nnx.nnx.nn import dtypes, initializers
from flax.typing import (
  Array,
  Dtype,
  Initializer,
  Axes,
)


def _canonicalize_axes(rank: int, axes: Axes) -> tp.Tuple[int, ...]:
  """Returns a tuple of deduplicated, sorted, and positive axes."""
  if not isinstance(axes, tp.Iterable):
    axes = (axes,)
  return tuple(set([rank + axis if axis < 0 else axis for axis in axes]))


def _abs_sq(x):
  """Computes the elementwise square of the absolute value |x|^2."""
  if jnp.iscomplexobj(x):
    return lax.square(lax.real(x)) + lax.square(lax.imag(x))
  else:
    return lax.square(x)


def _compute_stats(
  x: Array,
  axes: Axes,
  dtype: tp.Optional[Dtype],
  axis_name: tp.Optional[str] = None,
  axis_index_groups: tp.Any = None,
  use_mean: bool = True,
  use_fast_variance: bool = True,
  mask: tp.Optional[Array] = None,
):
  """Computes mean and variance statistics.

  This implementation takes care of a few important details:
  - Computes in float32 precision for stability in half precision training.
  - If ``use_fast_variance`` is ``True``, mean and variance are computed using
    Var = E[|x|^2] - |E[x]|^2, instead of Var = E[|x - E[x]|^2]), in a single
    XLA fusion.
  - Clips negative variances to zero which can happen due to
    roundoff errors. This avoids downstream NaNs.
  - Supports averaging across a parallel axis and subgroups of a parallel axis
    with a single ``lax.pmean`` call to avoid latency.

  Arguments:
    x: Input array.
    axes: The axes in ``x`` to compute mean and variance statistics for.
    dtype: Optional dtype specifying the minimal precision. Statistics are
      always at least float32 for stability (default: dtype of x).
    axis_name: Optional name for the pmapped axis to compute mean over. Note,
      this is only used for pmap and shard map. For SPMD jit, you do not need to
      manually synchronize. Just make sure that the axes are correctly annotated
      and XLA:SPMD will insert the necessary collectives.
    axis_index_groups: Optional axis indices.
    use_mean: If true, calculate the mean from the input and use it when
      computing the variance. If false, set the mean to zero and compute the
      variance without subtracting the mean.
    use_fast_variance: If true, use a faster, but less numerically stable,
      calculation for the variance.
    mask: Binary array of shape broadcastable to ``inputs`` tensor, indicating
      the positions for which the mean and variance should be computed.

  Returns:
    A pair ``(mean, var)``.
  """
  if dtype is None:
    dtype = jnp.result_type(x)
  # promote x to at least float32, this avoids half precision computation
  # but preserves double or complex floating points
  dtype = jnp.promote_types(dtype, jnp.float32)
  x = jnp.asarray(x, dtype)
  axes = _canonicalize_axes(x.ndim, axes)

  def maybe_distributed_mean(*xs, mask=None):
    mus = tuple(x.mean(axes, where=mask) for x in xs)
    if axis_name is None:
      return mus if len(xs) > 1 else mus[0]
    else:
      # In the distributed case we stack multiple arrays to speed comms.
      if len(xs) > 1:
        reduced_mus = lax.pmean(
          jnp.stack(mus, axis=0),
          axis_name,
          axis_index_groups=axis_index_groups,
        )
        return tuple(reduced_mus[i] for i in range(len(xs)))
      else:
        return lax.pmean(mus[0], axis_name, axis_index_groups=axis_index_groups)

  if use_mean:
    if use_fast_variance:
      mu, mu2 = maybe_distributed_mean(x, _abs_sq(x), mask=mask)
      # mean2 - _abs_sq(mean) is not guaranteed to be non-negative due
      # to floating point round-off errors.
      var = jnp.maximum(0.0, mu2 - _abs_sq(mu))
    else:
      mu = maybe_distributed_mean(x, mask=mask)
      var = maybe_distributed_mean(
        _abs_sq(x - jnp.expand_dims(mu, axes)), mask=mask
      )
  else:
    var = maybe_distributed_mean(_abs_sq(x), mask=mask)
    mu = jnp.zeros_like(var)
  return mu, var


def _normalize(
  x: Array,
  mean: Array,
  var: Array,
  scale: tp.Optional[Array],
  bias: tp.Optional[Array],
  reduction_axes: Axes,
  feature_axes: Axes,
  dtype: tp.Optional[Dtype],
  epsilon: float,
):
  """ "Normalizes the input of a normalization layer and optionally applies a learned scale and bias.

  Arguments:
    x: The input.
    mean: Mean to use for normalization.
    var: Variance to use for normalization.
    reduction_axes: The axes in ``x`` to reduce.
    feature_axes: Axes containing features. A separate bias and scale is learned
      for each specified feature.
    dtype: The dtype of the result (default: infer from input and params).
    epsilon: Normalization epsilon.

  Returns:
    The normalized input.
  """
  reduction_axes = _canonicalize_axes(x.ndim, reduction_axes)
  feature_axes = _canonicalize_axes(x.ndim, feature_axes)
  stats_shape = list(x.shape)
  for axis in reduction_axes:
    stats_shape[axis] = 1
  mean = mean.reshape(stats_shape)
  var = var.reshape(stats_shape)
  feature_shape = [1] * x.ndim
  reduced_feature_shape = []
  for ax in feature_axes:
    feature_shape[ax] = x.shape[ax]
    reduced_feature_shape.append(x.shape[ax])
  y = x - mean
  mul = lax.rsqrt(var + epsilon)
  args = [x]
  if scale is not None:
    scale = scale.reshape(feature_shape)
    mul *= scale
    args.append(scale)
  y *= mul
  if bias is not None:
    bias = bias.reshape(feature_shape)
    y += bias
    args.append(bias)
  dtype = dtypes.canonicalize_dtype(*args, dtype=dtype)
  return jnp.asarray(y, dtype)


class BatchNorm(Module):
  """BatchNorm Module.

  To calculate the batch norm on the input and update the batch statistics,
  call the :func:`train` method (or pass in ``use_running_average=False`` in
  the constructor or during call time).

  To use the stored batch statistics' running average, call the :func:`eval`
  method (or pass in ``use_running_average=True`` in the constructor or
  during call time).

  Example usage::

    >>> from flax import nnx
    >>> import jax, jax.numpy as jnp

    >>> x = jax.random.normal(jax.random.key(0), (5, 6))
    >>> layer = nnx.BatchNorm(num_features=6, momentum=0.9, epsilon=1e-5,
    ...                       dtype=jnp.float32, rngs=nnx.Rngs(0))
    >>> jax.tree.map(jnp.shape, nnx.state(layer))
    State({
      'bias': VariableState(
        type=Param,
        value=(6,)
      ),
      'mean': VariableState(
        type=BatchStat,
        value=(6,)
      ),
      'scale': VariableState(
        type=Param,
        value=(6,)
      ),
      'var': VariableState(
        type=BatchStat,
        value=(6,)
      )
    })

    >>> # calculate batch norm on input and update batch statistics
    >>> layer.train()
    >>> y = layer(x)
    >>> batch_stats1 = nnx.state(layer, nnx.BatchStat)
    >>> y = layer(x)
    >>> batch_stats2 = nnx.state(layer, nnx.BatchStat)
    >>> assert (batch_stats1['mean'].value != batch_stats2['mean'].value).all()
    >>> assert (batch_stats1['var'].value != batch_stats2['var'].value).all()

    >>> # use stored batch statistics' running average
    >>> layer.eval()
    >>> y = layer(x)
    >>> batch_stats3 = nnx.state(layer, nnx.BatchStat)
    >>> assert (batch_stats2['mean'].value == batch_stats3['mean'].value).all()
    >>> assert (batch_stats2['var'].value == batch_stats3['var'].value).all()

  Attributes:
    num_features: the number of input features.
    use_running_average: if True, the stored batch statistics will be
      used instead of computing the batch statistics on the input.
    axis: the feature or non-batch axis of the input.
    momentum: decay rate for the exponential moving average of
      the batch statistics.
    epsilon: a small float added to variance to avoid dividing by zero.
    dtype: the dtype of the result (default: infer from input and params).
    param_dtype: the dtype passed to parameter initializers (default: float32).
    use_bias:  if True, bias (beta) is added.
    use_scale: if True, multiply by scale (gamma).
      When the next layer is linear (also e.g. nn.relu), this can be disabled
      since the scaling will be done by the next layer.
    bias_init: initializer for bias, by default, zero.
    scale_init: initializer for scale, by default, one.
    axis_name: the axis name used to combine batch statistics from multiple
      devices. See ``jax.pmap`` for a description of axis names (default: None).
    axis_index_groups: groups of axis indices within that named axis
      representing subsets of devices to reduce over (default: None). For
      example, ``[[0, 1], [2, 3]]`` would independently batch-normalize over
      the examples on the first two and last two devices. See ``jax.lax.psum``
      for more details.
    use_fast_variance: If true, use a faster, but less numerically stable,
      calculation for the variance.
    rngs: rng key.
  """

  def __init__(
    self,
    num_features: int,
    *,
    use_running_average: bool = False,
    axis: int = -1,
    momentum: float = 0.99,
    epsilon: float = 1e-5,
    dtype: tp.Optional[Dtype] = None,
    param_dtype: Dtype = jnp.float32,
    use_bias: bool = True,
    use_scale: bool = True,
    bias_init: Initializer = initializers.zeros_init(),
    scale_init: Initializer = initializers.ones_init(),
    axis_name: tp.Optional[str] = None,
    axis_index_groups: tp.Any = None,
    use_fast_variance: bool = True,
    rngs: rnglib.Rngs,
  ):
    feature_shape = (num_features,)
    self.mean = nnx.BatchStat(jnp.zeros(feature_shape, jnp.float32))
    self.var = nnx.BatchStat(jnp.ones(feature_shape, jnp.float32))

    if use_scale:
      key = rngs.params()
      self.scale = nnx.Param(scale_init(key, feature_shape, param_dtype))
    else:
      self.scale = nnx.Param(None)

    if use_bias:
      key = rngs.params()
      self.bias = nnx.Param(bias_init(key, feature_shape, param_dtype))
    else:
      self.bias = nnx.Param(None)

    self.num_features = num_features
    self.use_running_average = use_running_average
    self.axis = axis
    self.momentum = momentum
    self.epsilon = epsilon
    self.dtype = dtype
    self.param_dtype = param_dtype
    self.use_bias = use_bias
    self.use_scale = use_scale
    self.bias_init = bias_init
    self.scale_init = scale_init
    self.axis_name = axis_name
    self.axis_index_groups = axis_index_groups
    self.use_fast_variance = use_fast_variance

  def __call__(
    self,
    x,
    use_running_average: tp.Optional[bool] = None,
    *,
    mask: tp.Optional[jax.Array] = None,
  ):
    """Normalizes the input using batch statistics.

    Args:
      x: the input to be normalized.
      use_running_average: if true, the stored batch statistics will be
        used instead of computing the batch statistics on the input. The
        ``use_running_average`` flag passed into the call method will take
        precedence over the ``use_running_average`` flag passed into the
        constructor.

    Returns:
      Normalized inputs (the same shape as inputs).
    """

    use_running_average = first_from(
      use_running_average,
      self.use_running_average,
      error_msg="""No `use_running_average` argument was provided to BatchNorm
        as either a __call__ argument, class attribute, or nnx.flag.""",
    )
    feature_axes = _canonicalize_axes(x.ndim, self.axis)
    reduction_axes = tuple(i for i in range(x.ndim) if i not in feature_axes)

    if use_running_average:
      mean, var = self.mean.value, self.var.value
    else:
      mean, var = _compute_stats(
        x,
        reduction_axes,
        dtype=self.dtype,
        axis_name=self.axis_name,
        axis_index_groups=self.axis_index_groups,
        use_fast_variance=self.use_fast_variance,
        mask=mask,
      )

      self.mean.value = (
        self.momentum * self.mean.value + (1 - self.momentum) * mean
      )
      self.var.value = (
        self.momentum * self.var.value + (1 - self.momentum) * var
      )

    return _normalize(
      x,
      mean,
      var,
      self.scale.value,
      self.bias.value,
      reduction_axes,
      feature_axes,
      self.dtype,
      self.epsilon,
    )


class LayerNorm(Module):
  """Layer normalization (https://arxiv.org/abs/1607.06450).

  LayerNorm normalizes the activations of the layer for each given example in a
  batch independently, rather than across a batch like Batch Normalization.
  i.e. applies a transformation that maintains the mean activation within
  each example close to 0 and the activation standard deviation close to 1.

  Example usage::

    >>> from flax import nnx
    >>> import jax

    >>> x = jax.random.normal(jax.random.key(0), (3, 4, 5, 6))
    >>> layer = nnx.LayerNorm(num_features=6, rngs=nnx.Rngs(0))

    >>> nnx.state(layer)
    State({
      'bias': VariableState(
        type=Param,
        value=Array([0., 0., 0., 0., 0., 0.], dtype=float32)
      ),
      'scale': VariableState(
        type=Param,
        value=Array([1., 1., 1., 1., 1., 1.], dtype=float32)
      )
    })

    >>> y = layer(x)

  Attributes:
    num_features: the number of input features.
    epsilon: A small float added to variance to avoid dividing by zero.
    dtype: the dtype of the result (default: infer from input and params).
    param_dtype: the dtype passed to parameter initializers (default: float32).
    use_bias:  If True, bias (beta) is added.
    use_scale: If True, multiply by scale (gamma). When the next layer is linear
        (also e.g. nnx.relu), this can be disabled since the scaling will be done
        by the next layer.
    bias_init: Initializer for bias, by default, zero.
    scale_init: Initializer for scale, by default, one.
    reduction_axes: Axes for computing normalization statistics.
    feature_axes: Feature axes for learned bias and scaling.
    axis_name: the axis name used to combine batch statistics from multiple
        devices. See ``jax.pmap`` for a description of axis names (default: None).
        This is only needed if the model is subdivided across devices, i.e. the
        array being normalized is sharded across devices within a pmap.
    axis_index_groups: groups of axis indices within that named axis
        representing subsets of devices to reduce over (default: None). For
        example, ``[[0, 1], [2, 3]]`` would independently batch-normalize over
        the examples on the first two and last two devices. See ``jax.lax.psum``
        for more details.
    use_fast_variance: If true, use a faster, but less numerically stable,
        calculation for the variance.
    rngs: rng key.
  """

  def __init__(
    self,
    num_features: int,
    *,
    epsilon: float = 1e-6,
    dtype: tp.Optional[Dtype] = None,
    param_dtype: Dtype = jnp.float32,
    use_bias: bool = True,
    use_scale: bool = True,
    bias_init: Initializer = initializers.zeros_init(),
    scale_init: Initializer = initializers.ones_init(),
    reduction_axes: Axes = -1,
    feature_axes: Axes = -1,
    axis_name: tp.Optional[str] = None,
    axis_index_groups: tp.Any = None,
    use_fast_variance: bool = True,
    rngs: rnglib.Rngs,
  ):
    feature_shape = (num_features,)

    if use_scale:
      key = rngs.params()
      self.scale = nnx.Param(scale_init(key, feature_shape, param_dtype))
    else:
      self.scale = nnx.Param(None)

    if use_bias:
      key = rngs.params()
      self.bias = nnx.Param(bias_init(key, feature_shape, param_dtype))
    else:
      self.bias = nnx.Param(None)

    self.num_features = num_features
    self.epsilon = epsilon
    self.dtype = dtype
    self.param_dtype = param_dtype
    self.use_bias = use_bias
    self.use_scale = use_scale
    self.bias_init = bias_init
    self.scale_init = scale_init
    self.reduction_axes = reduction_axes
    self.feature_axes = feature_axes
    self.axis_name = axis_name
    self.axis_index_groups = axis_index_groups
    self.use_fast_variance = use_fast_variance

  def __call__(self, x, *, mask: tp.Optional[jax.Array] = None):
    """Applies layer normalization on the input.

    Args:
      x: the inputs

    Returns:
      Normalized inputs (the same shape as inputs).
    """
    mean, var = _compute_stats(
      x,
      self.reduction_axes,
      self.dtype,
      self.axis_name,
      self.axis_index_groups,
      use_fast_variance=self.use_fast_variance,
      mask=mask,
    )

    return _normalize(
      x,
      mean,
      var,
      self.scale.value,
      self.bias.value,
      self.reduction_axes,
      self.feature_axes,
      self.dtype,
      self.epsilon,
    )


class RMSNorm(Module):
  """RMS Layer normalization (https://arxiv.org/abs/1910.07467).

  RMSNorm normalizes the activations of the layer for each given example in a
  batch independently, rather than across a batch like Batch Normalization.
  Unlike LayerNorm which re-centers the mean to be 0 and normalizes by the
  standard deviation of the activations, RMSNorm does not re-center at all
  and instead normalizes by the root mean square of the activations.

  Example usage::

    >>> from flax import nnx
    >>> import jax

    >>> x = jax.random.normal(jax.random.key(0), (5, 6))
    >>> layer = nnx.RMSNorm(num_features=6, rngs=nnx.Rngs(0))

    >>> nnx.state(layer)
    State({
      'scale': VariableState(
        type=Param,
        value=Array([1., 1., 1., 1., 1., 1.], dtype=float32)
      )
    })

    >>> y = layer(x)

  Attributes:
    num_features: the number of input features.
    epsilon: A small float added to variance to avoid dividing by zero.
    dtype: the dtype of the result (default: infer from input and params).
    param_dtype: the dtype passed to parameter initializers (default: float32).
    use_scale: If True, multiply by scale (gamma). When the next layer is linear
        (also e.g. nn.relu), this can be disabled since the scaling will be done
        by the next layer.
    scale_init: Initializer for scale, by default, one.
    reduction_axes: Axes for computing normalization statistics.
    feature_axes: Feature axes for learned bias and scaling.
    axis_name: the axis name used to combine batch statistics from multiple
        devices. See ``jax.pmap`` for a description of axis names (default: None).
        This is only needed if the model is subdivided across devices, i.e. the
        array being normalized is sharded across devices within a pmap.
    axis_index_groups: groups of axis indices within that named axis
        representing subsets of devices to reduce over (default: None). For
        example, ``[[0, 1], [2, 3]]`` would independently batch-normalize over
        the examples on the first two and last two devices. See ``jax.lax.psum``
        for more details.
    use_fast_variance: If true, use a faster, but less numerically stable,
        calculation for the variance.
    rngs: rng key.
  """

  def __init__(
    self,
    num_features: int,
    *,
    epsilon: float = 1e-6,
    dtype: tp.Optional[Dtype] = None,
    param_dtype: Dtype = jnp.float32,
    use_scale: bool = True,
    scale_init: Initializer = initializers.ones,
    reduction_axes: Axes = -1,
    feature_axes: Axes = -1,
    axis_name: tp.Optional[str] = None,
    axis_index_groups: tp.Any = None,
    use_fast_variance: bool = True,
    rngs: rnglib.Rngs,
  ):
    feature_shape = (num_features,)

    if use_scale:
      key = rngs.params()
      self.scale = nnx.Param(scale_init(key, feature_shape, param_dtype))
    else:
      self.scale = nnx.Param(None)

    self.num_features = num_features
    self.epsilon = epsilon
    self.dtype = dtype
    self.param_dtype = param_dtype
    self.use_scale = use_scale
    self.scale_init = scale_init
    self.reduction_axes = reduction_axes
    self.feature_axes = feature_axes
    self.axis_name = axis_name
    self.axis_index_groups = axis_index_groups
    self.use_fast_variance = use_fast_variance

  def __call__(self, x, mask: tp.Optional[jax.Array] = None):
    """Applies layer normalization on the input.

    Args:
      x: the inputs

    Returns:
      Normalized inputs (the same shape as inputs).
    """
    mean, var = _compute_stats(
      x,
      self.reduction_axes,
      self.dtype,
      self.axis_name,
      self.axis_index_groups,
      use_mean=False,
      use_fast_variance=self.use_fast_variance,
      mask=mask,
    )

    return _normalize(
      x,
      mean,
      var,
      self.scale.value,
      None,
      self.reduction_axes,
      self.feature_axes,
      self.dtype,
      self.epsilon,
    )



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import tempfile

from absl.testing import absltest
import tensorflow_datasets as tfds

import input_pipeline
import vocabulary


class InputPipelineTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.vocab_path = self._get_vocab_path()
    self.dataset = self._get_dataset(self.vocab_path)

  def _get_vocab_path(self):
    """Prepares a mock vocabulary and returns a path to it."""
    vocab_path = os.path.join(tempfile.mkdtemp(), 'vocab.txt')
    tokenized_sequences = (
        (b'this', b'is', b'a', b'test', b'sentence'),
        (b'this', b'is', b'a', b'test', b'sentence'),
        (b'this', b'is', b'a', b'test', b'sentence'),
    )
    vocab = vocabulary.Vocabulary(tokenized_sequences=tokenized_sequences)
    vocab.save(vocab_path)
    return vocab_path

  def _get_dataset(self, vocab_path):
    """Uses mock data to create the dataset."""
    # Go two directories up to the root of the flax directory.
    flax_root_dir = pathlib.Path(__file__).parents[2]
    data_dir = str(flax_root_dir) + '/.tfds/metadata'  # pylint: disable=unused-variable
    with tfds.testing.mock_data(num_examples=128, data_dir=data_dir):
      return input_pipeline.TextDataset(vocab_path=vocab_path, split='train')

  def test_bucketed_dataset(self):
    """Each batch should have a length that is a multiple of bucket_size."""
    batch_size = 2
    bucket_size = 8
    for batch in self.dataset.get_bucketed_batches(
        batch_size=batch_size,
        bucket_size=bucket_size,
        max_input_length=60,
        shuffle=False,
    ).take(3):
      # Because of bucketing, sequence length must be multiple of bucket_size.
      length = batch['token_ids'].numpy().shape[-1]
      self.assertEqual(0, length % bucket_size)
      self.assertEqual((batch_size,), batch['length'].numpy().shape)
      self.assertEqual((batch_size,), batch['label'].numpy().shape)

  def test_batched_dataset(self):
    """Tests that the length of a batch matches the longest sequence."""
    batch_size = 2
    for batch in self.dataset.get_batches(
        batch_size=batch_size, shuffle=False
    ).take(1):
      # Each batch is padded to the maximum sentence length in the batch.
      max_length_in_batch = max(batch['length'].numpy())
      length = batch['token_ids'].numpy().shape[-1]
      self.assertEqual(max_length_in_batch, length)
      self.assertEqual((batch_size,), batch['length'].numpy().shape)
      self.assertEqual((batch_size,), batch['label'].numpy().shape)

  def test_batched_dataset_fixed_length(self):
    """Tests that each batch has the fixed length."""
    batch_size = 2
    fixed_pad_length = 77
    for batch in self.dataset.get_batches(
        batch_size=batch_size, shuffle=False, fixed_pad_length=fixed_pad_length
    ).take(1):
      length = batch['token_ids'].numpy().shape[-1]
      self.assertEqual(fixed_pad_length, length)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Sequence

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random
from jax.scipy.linalg import expm

from flax.core import Array, Scope, apply, init, nn, unfreeze

Initializer = Any
Flow = Any


@dataclass
class DenseFlow:
  kernel_init: Initializer = nn.linear.default_kernel_init
  bias_init: Initializer = nn.initializers.zeros_init()

  def params(self, scope: Scope, features: int):
    kernel = scope.param('kernel', self.kernel_init, (features, features))
    bias = scope.param('bias', self.bias_init, (features,))
    return kernel, bias

  def forward(self, scope: Scope, x: Array):
    kernel, bias = self.params(scope, x.shape[-1])
    return jnp.dot(x, expm(kernel)) + bias.reshape((1,) * (x.ndim - 1) + (-1,))

  def backward(self, scope: Scope, y: Array):
    kernel, bias = self.params(scope, y.shape[-1])
    return jnp.dot(y - bias.reshape((1,) * (y.ndim - 1) + (-1,)), expm(-kernel))


@dataclass
class StackFlow:
  flows: Sequence[Flow]

  def forward(self, scope: Scope, x: Array):
    for i, f in enumerate(self.flows):
      x = scope.child(f.forward, name=str(i))(x)
    return x

  def backward(self, scope: Scope, x: Array):
    for i, f in reversed(tuple(enumerate(self.flows))):
      x = scope.child(f.backward, name=str(i))(x)
    return x


class FlowTest(absltest.TestCase):
  def test_flow(self):
    x = jnp.ones((1, 3))
    flow = StackFlow((DenseFlow(),) * 3)
    y, variables = init(flow.forward)(random.key(0), x)
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(y.shape, (1, 3))
    self.assertEqual(
      param_shapes,
      {
        '0': {'kernel': (3, 3), 'bias': (3,)},
        '1': {'kernel': (3, 3), 'bias': (3,)},
        '2': {'kernel': (3, 3), 'bias': (3,)},
      },
    )
    x_restored = apply(flow.backward)(variables, y)
    self.assertTrue(jnp.allclose(x, x_restored))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

from jax.nn.initializers import constant as constant
from jax.nn.initializers import delta_orthogonal as delta_orthogonal
from jax.nn.initializers import glorot_normal as glorot_normal
from jax.nn.initializers import glorot_uniform as glorot_uniform
from jax.nn.initializers import he_normal as he_normal
from jax.nn.initializers import he_uniform as he_uniform
from jax.nn.initializers import kaiming_normal as kaiming_normal
from jax.nn.initializers import kaiming_uniform as kaiming_uniform
from jax.nn.initializers import lecun_normal as lecun_normal
from jax.nn.initializers import lecun_uniform as lecun_uniform
from jax.nn.initializers import normal as normal
from jax.nn.initializers import ones as ones
from jax.nn.initializers import orthogonal as orthogonal
from jax.nn.initializers import truncated_normal as truncated_normal
from jax.nn.initializers import uniform as uniform
from jax.nn.initializers import variance_scaling as variance_scaling
from jax.nn.initializers import xavier_normal as xavier_normal
from jax.nn.initializers import xavier_uniform as xavier_uniform
from jax.nn.initializers import zeros as zeros
from flax.typing import Initializer

DtypeLikeInexact = tp.Any


def zeros_init() -> Initializer:
  """Builds an initializer that returns a constant array full of zeros.

  >>> import jax, jax.numpy as jnp
  >>> from flax.nnx import initializers
  >>> zeros_initializer = initializers.zeros_init()
  >>> zeros_initializer(jax.random.key(42), (2, 3), jnp.float32)
  Array([[0., 0., 0.],
         [0., 0., 0.]], dtype=float32)
  """
  return zeros


def ones_init() -> Initializer:
  """Builds an initializer that returns a constant array full of ones.

  >>> import jax, jax.numpy as jnp
  >>> from flax.nnx import initializers
  >>> ones_initializer = initializers.ones_init()
  >>> ones_initializer(jax.random.key(42), (3, 2), jnp.float32)
  Array([[1., 1.],
         [1., 1.],
         [1., 1.]], dtype=float32)
  """
  return ones



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.linen.attention."""

from absl.testing import absltest, parameterized
from flax import errors, jax_utils
from flax import linen as nn
from flax.core import pop
import jax
from jax import lax, random
from jax.nn import initializers
import jax.numpy as jnp
import numpy as np

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class AttentionTest(parameterized.TestCase):
  def test_multihead_self_attention(self):
    rng = random.key(0)
    x = jnp.ones((4, 6, 5))
    sa_module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      deterministic=False,
    )
    y, _ = sa_module.init_with_output(rng, x)
    self.assertEqual(y.shape, x.shape)
    self.assertEqual(y.dtype, jnp.float32)

  def test_dtype_infer(self):
    rng = random.key(0)
    x = jnp.ones((4, 6, 5), jnp.complex64)
    sa_module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      deterministic=False,
    )
    y, _ = sa_module.init_with_output(rng, x)
    self.assertEqual(y.shape, x.shape)
    self.assertEqual(y.dtype, jnp.complex64)

  def test_multihead_encoder_decoder_attention(self):
    rng = random.key(0)
    q = jnp.ones((4, 2, 3, 5))
    sa_module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      deterministic=False,
    )
    y, _ = sa_module.init_with_output(rng, q)
    self.assertEqual(y.shape, q.shape)

  def test_mha_out_initializers(self):
    rng = random.key(0)
    q = jnp.ones((4, 2, 3, 5))
    sa_module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      out_kernel_init=initializers.zeros,
      bias_init=initializers.zeros,
      out_bias_init=initializers.ones,
      deterministic=False,
    )
    variables = sa_module.init(rng, q)
    params = variables['params']
    # test kernels
    np.testing.assert_allclose(params['query']['kernel'], 1.0)
    np.testing.assert_allclose(params['key']['kernel'], 1.0)
    np.testing.assert_allclose(params['value']['kernel'], 1.0)
    np.testing.assert_allclose(params['out']['kernel'], 0.0)
    # test biases
    np.testing.assert_allclose(params['query']['bias'], 0.0)
    np.testing.assert_allclose(params['key']['bias'], 0.0)
    np.testing.assert_allclose(params['value']['bias'], 0.0)
    np.testing.assert_allclose(params['out']['bias'], 1.0)

  def test_multihead_self_attention_w_dropout(self):
    rng = random.key(0)
    x = jnp.ones((4, 2, 3, 5))
    sa_module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      dropout_rate=0.1,
      deterministic=False,
    )
    rng1, rng2 = random.split(rng)
    rngs = {'params': rng1, 'dropout': rng2}
    y, _ = sa_module.init_with_output(rngs, x)
    self.assertEqual(y.shape, x.shape)

  def test_multihead_self_attention_explicit_dropout(self):
    def clone(key):
      if hasattr(jax.random, "clone"):
        # JAX v0.4.26+
        return jax.tree.map(jax.random.clone, key)
      return key

    class Foo(nn.Module):
      attention_kwargs: dict

      @nn.compact
      def __call__(self, x, dropout_rng=None):
        a = nn.MultiHeadDotProductAttention(**self.attention_kwargs)(
          x, x, dropout_rng=dropout_rng
        )
        if dropout_rng is not None:
          dropout_rng = clone(dropout_rng)
        b = nn.MultiHeadDotProductAttention(**self.attention_kwargs)(
          x, x, dropout_rng=dropout_rng
        )
        return a, b

    module = Foo(
      dict(
        num_heads=8,
        qkv_features=16,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        dropout_rate=0.5,
        deterministic=False,
      )
    )
    rng1, rng2, rng3, rng4 = random.split(random.key(0), 4)
    x = jnp.ones((4, 2, 3, 5))
    rngs = {'params': rng1, 'dropout': rng2}
    v = module.init(rngs, x)
    a, b = module.apply(v, x, rngs=clone(rngs))
    c, d = module.apply(v, x, rngs={'dropout': clone(rng2)})
    e, f = module.apply(v, x, rngs={'dropout': rng3})
    self.assertFalse((a == b).all())
    self.assertTrue((a == c).all())
    self.assertTrue((b == d).all())
    self.assertFalse((a == e).all())
    self.assertFalse((b == f).all())
    a, b = module.apply(v, x, rngs=clone(rngs), dropout_rng=rng4)
    self.assertTrue((a == b).all())
    a, b = module.apply(v, x, dropout_rng=clone(rng4))
    self.assertTrue((a == b).all())
    self.assertTrue(a.shape == b.shape == x.shape)

  def test_multihead_self_attention_w_dropout_disabled(self):
    rng = random.key(0)
    x = jnp.ones((4, 2, 3, 5))
    sa_module0 = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      dropout_rate=0.0,
      deterministic=True,
    )
    rng1, rng2, rng3, rng4 = random.split(rng, 4)
    rngs1 = {'params': rng1, 'dropout': rng2}
    rngs2 = {'params': rng3, 'dropout': rng4}
    y1, vs = sa_module0.init_with_output(rngs1, x)
    y2, _ = sa_module0.init_with_output(rngs2, x)
    np.testing.assert_allclose(y1, y2)
    y3 = sa_module0.apply(vs, x, rngs=rngs1)
    y4 = sa_module0.apply(vs, x, rngs=rngs2)
    np.testing.assert_allclose(y3, y4)
    sa_module1 = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      dropout_rate=0.0,
    )
    y5 = sa_module1.apply(vs, x, deterministic=True, rngs=rngs1)
    y6 = sa_module1.apply(vs, x, deterministic=True, rngs=rngs2)
    np.testing.assert_allclose(y5, y6)
    sa_module2 = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      dropout_rate=0.5,
    )
    y7 = sa_module2.apply(vs, x, deterministic=True, rngs=rngs1)
    y8 = sa_module2.apply(vs, x, deterministic=True, rngs=rngs2)
    np.testing.assert_allclose(y7, y8)

  def test_causal_mask_1d(self):
    """Tests autoregresive masking for 1d attention."""
    x = jnp.ones((3, 16))  # (bs1, length)
    mask_1d = nn.attention.make_causal_mask(x)
    ts = np.arange(16)
    mask_1d_simple = (ts[:, None] >= ts[None, :])[None, None, :, :]
    mask_1d_simple = jnp.broadcast_to(mask_1d_simple, (3, 1, 16, 16))
    np.testing.assert_allclose(
      mask_1d,
      mask_1d_simple,
    )

  @parameterized.parameters([((5,), (1,)), ((6, 5), (2,))])
  def test_decoding(self, spatial_shape, attn_dims):
    bs = 2
    num_heads = 3
    num_features = 4
    rng = random.key(0)
    key1, key2 = random.split(rng)
    inputs = random.normal(
      key1, (bs,) + spatial_shape + (num_heads * num_features,)
    )
    module = nn.MultiHeadDotProductAttention(
      num_heads=num_heads,
      qkv_features=num_heads * num_features,
      precision=lax.Precision.HIGHEST,
      deterministic=False,
      decode=False,
    )
    decode_module = module.clone(decode=True)

    initial_vars = decode_module.init(key2, inputs)
    state, params = pop(initial_vars, 'params')
    causal_mask = nn.attention.make_causal_mask(jnp.ones((bs,) + spatial_shape))
    y_ref = jax.jit(lambda x, y: module.apply(initial_vars, x, mask=y))(
      inputs, causal_mask
    )

    # feed the inputs sequentially to simulate decoding
    def body_fn(state, x):
      y, state = decode_module.apply(
        {'params': params, **state}, x, mutable=['cache']
      )
      return state, y

    # scan_in_dim supports scanning multiple dims
    _, y = jax_utils.scan_in_dim(
      body_fn, state, inputs, axis=attn_dims, keepdims=True
    )

    np.testing.assert_allclose(y_ref, y, atol=1e-5)

  def test_autoregresive_receptive_field_1d(self):
    """Tests the autoregresive self-attention receptive field."""
    rng = random.key(0)
    rng1, rng2 = random.split(rng, num=2)

    length = 10
    dim = 1
    num_heads = 1
    input_shape = (1, length, dim)
    inputs = random.normal(rng2, input_shape)

    module = nn.MultiHeadDotProductAttention(
      num_heads=num_heads,
      kernel_init=jax.nn.initializers.ones,
      deterministic=False,
    )

    initial_vars = module.init(rng1, inputs)
    causal_mask = nn.attention.make_causal_mask(jnp.ones(input_shape[:-1]))

    def model_loss(inputs, pos):
      out = module.apply(initial_vars, inputs, mask=causal_mask)
      assert out.shape == input_shape
      assert len(out.shape) == 3
      return out[0, pos, :].sum()

    grad_fn = jax.jit(jax.grad(model_loss))

    def get_receptive_field_1d(pos):
      g = grad_fn(inputs, pos)[0, :, :]
      return jnp.any((jnp.abs(g) > 1e-5).astype(jnp.uint32), axis=-1)

    for i in range(length):
      deps = get_receptive_field_1d(i)
      assert (deps[:i] == 1).all(), (
        'Receptive Field Error: Some of the '
        'previous postions are not reachable '
        'in autoregressive self-attention.'
      )
      if i != length - 1:
        k = i + 1
        assert (deps[k:] == 0).all(), (
          'Receptive Field Error: Some of the '
          'future postions are reachable in '
          'autoregressive self-attention.'
        )

  def test_multihead_kv_args(self):
    key1, key2 = random.split(random.key(0), 2)
    query = random.uniform(key1, (3, 5))
    key_value = random.uniform(key2, (9, 5))
    module = nn.MultiHeadDotProductAttention(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.ones,
      bias_init=initializers.zeros,
      deterministic=False,
    )
    key = lambda: random.key(43279)
    y0, v0 = module.init_with_output(
      key(), query, inputs_k=key_value, inputs_v=key_value
    )
    y1, v1 = module.init_with_output(key(), query, inputs_k=key_value)
    with self.assertWarnsRegex(
      DeprecationWarning, 'The inputs_kv arg will be deprecated soon.'
    ):
      y2, v2 = module.init_with_output(key(), query, inputs_kv=key_value)
    self.assertTrue((y0 == y1).all() and (y1 == y2).all())
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(
                lambda x, y, z: (x == y).all() and (y == z).all(), v0, v1, v2
            )
        )
    )

    with self.assertRaisesRegex(
      ValueError, '`inputs_k` cannot be None if `inputs_v` is not None.'
    ):
      y3, v3 = module.init_with_output(key(), query, inputs_v=key_value)
    with self.assertRaisesRegex(
      ValueError,
      'If either `inputs_k` or `inputs_v` is not None, `inputs_kv` must be None.',
    ):
      y3, v3 = module.init_with_output(
        key(), query, inputs_kv=key_value, inputs_v=key_value
      )
    with self.assertRaisesRegex(
      ValueError,
      'If either `inputs_k` or `inputs_v` is not None, `inputs_kv` must be None.',
    ):
      y3, v3 = module.init_with_output(
        key(), query, key_value, key_value, inputs_kv=key_value
      )

  def test_multihead_mask_warning(self):
    rng = random.key(0)
    rng1, rng2 = random.split(rng, num=2)

    length = 10
    dim = 1
    num_heads = 1
    input_shape = (1, length, dim)
    query = key = random.normal(rng2, input_shape)

    module = nn.MultiHeadDotProductAttention(
      num_heads=num_heads,
      kernel_init=jax.nn.initializers.ones,
      deterministic=False,
    )

    initial_vars = module.init(rng1, query, key)
    causal_mask = nn.attention.make_causal_mask(jnp.ones(input_shape[:-1]))

    module.apply(initial_vars, query, key, mask=causal_mask)
    with self.assertWarnsRegex(
      DeprecationWarning,
      "the function signature of MultiHeadDotProductAttention's `__call__` method has changed",
    ):
      with self.assertRaises(errors.ScopeParamShapeError):
        module.apply(initial_vars, query, key, causal_mask)

  def test_multihead_sow_attention_weights(self):
    rng = random.key(0)
    x = jnp.ones((4, 6, 5))

    class Model(nn.Module):
      attention_kwargs: dict

      @nn.compact
      def __call__(self, x, sow_weights=False):
        x = nn.MultiHeadDotProductAttention(**self.attention_kwargs)(
          x, sow_weights=sow_weights
        )
        x = nn.MultiHeadDotProductAttention(**self.attention_kwargs)(x)
        x = nn.MultiHeadDotProductAttention(**self.attention_kwargs)(
          x, sow_weights=sow_weights
        )
        return x

    module = Model(
      dict(
        num_heads=8,
        qkv_features=16,
        kernel_init=initializers.ones,
        bias_init=initializers.zeros,
        deterministic=False,
      )
    )
    v = module.init(rng, x)
    _, intermediates = module.apply(
      v, x, mutable=['intermediates'], sow_weights=True
    )
    self.assertEqual(
      intermediates['intermediates']['MultiHeadDotProductAttention_0'][
        'attention_weights'
      ][0].shape,
      (4, 8, 6, 6),
    )
    self.assertNotIn(
      'MultiHeadDotProductAttention_1', intermediates['intermediates']
    )
    self.assertEqual(
      intermediates['intermediates']['MultiHeadDotProductAttention_2'][
        'attention_weights'
      ][0].shape,
      (4, 8, 6, 6),
    )
    _, intermediates = module.apply(
      v, x, mutable=['intermediates'], sow_weights=False
    )
    self.assertNotIn('intermediates', intermediates)

  def test_autoregressive_decode_with_x64(self):
    with jax.experimental.enable_x64():
      x = jnp.ones((1, 4, 4))
      module = nn.MultiHeadDotProductAttention(
          num_heads=2,
          qkv_features=4,
          decode=True
        )

      rng = random.PRNGKey(0)
      variables = module.init(rng, x, x, x)
      params, cache = variables['params'], variables['cache']
      y1, updates = module.apply(
          { 'params': params, 'cache': cache },
          x[:, :1, :],
          mutable=['cache']
        )
      cache = updates['cache']
      y2, updates = module.apply(
          { 'params': params, 'cache': cache },
          x[:, 1:2, :],
          mutable=['cache']
        )
      assert y1.shape == (1, 1, 4)
      assert y2.shape == (1, 1, 4)

  def test_attention_alias_equivalence(self):
    key1, key2 = random.split(random.key(0), 2)
    query = random.uniform(key1, (3, 5))
    key_value = random.uniform(key2, (9, 5))
    attention_kwargs = dict(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.lecun_normal(),
      bias_init=initializers.uniform(),
      deterministic=False,
    )
    module1 = nn.MultiHeadDotProductAttention(**attention_kwargs)
    module2 = nn.MultiHeadAttention(**attention_kwargs)
    key = lambda: random.key(43279)
    out1, v1 = module1.init_with_output(key(), query, key_value)
    out2, v2 = module2.init_with_output(key(), query, key_value, key_value)
    self.assertTrue((out1 == out2).all())
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(lambda x, y: (x == y).all(), v1, v2)
        )
    )

  def test_attention_alias_submodule(self):
    key1, key2 = random.split(random.key(0), 2)
    query = random.uniform(key1, (3, 5))
    key_value = random.uniform(key2, (9, 5))
    attention_kwargs = dict(
      num_heads=8,
      qkv_features=16,
      kernel_init=initializers.lecun_normal(),
      bias_init=initializers.uniform(),
      deterministic=False,
    )

    class Foo1(nn.Module):
      attention_kwargs: dict

      @nn.compact
      def __call__(self, query, key):
        return nn.MultiHeadDotProductAttention(**self.attention_kwargs)(
          query, key
        )

    class Foo2(nn.Module):
      attention_kwargs: dict

      @nn.compact
      def __call__(self, query, key, value):
        return nn.MultiHeadAttention(**self.attention_kwargs)(query, key, value)

    key = lambda: random.key(5478392)
    module1 = Foo1(attention_kwargs)
    module2 = Foo2(attention_kwargs)
    out1, v1 = module1.init_with_output(key(), query, key_value)
    out2, v2 = module2.init_with_output(key(), query, key_value, key_value)

    # test different output and variables if layer names are different
    self.assertTrue((out1 != out2).all())
    v2['params']['MultiHeadDotProductAttention_0'] = v2['params'][
      'MultiHeadAttention_0'
    ]
    del v2['params']['MultiHeadAttention_0']
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(lambda x, y: (x != y).all(), v1, v2)
        )
    )

    # test same output if variables are the same
    v2 = jax.tree_util.tree_map(lambda x: x, v1)
    v2['params']['MultiHeadAttention_0'] = v2['params'][
      'MultiHeadDotProductAttention_0'
    ]
    del v2['params']['MultiHeadDotProductAttention_0']
    out2 = module2.apply(v2, query, key_value, key_value)
    self.assertTrue((out1 == out2).all())

    # test same output and variables if names are the same
    class Foo2(nn.Module):
      attention_kwargs: dict

      @nn.compact
      def __call__(self, query, key, value):
        return nn.MultiHeadAttention(
          **self.attention_kwargs, name='MultiHeadDotProductAttention_0'
        )(query, key, value)

    module2 = Foo2(attention_kwargs)
    out2, v2 = module2.init_with_output(key(), query, key_value, key_value)
    self.assertTrue((out1 == out2).all())
    self.assertTrue(
        jax.tree_util.tree_all(
            jax.tree_util.tree_map(lambda x, y: (x == y).all(), v1, v2)
        )
    )

  @parameterized.parameters(
      {'force_fp32': True, 'attn_weights_dtype': jnp.float32},
      {'force_fp32': False, 'attn_weights_dtype': jnp.bfloat16},
  )
  def test_mixed_precision_multihead_attention(
      self, force_fp32, attn_weights_dtype
  ):
    input_key, params_key, dropout_key = random.split(random.key(0), 3)
    x = random.uniform(input_key, (2, 4))
    attention_kwargs = dict(
        num_heads=2,
        qkv_features=4,
        kernel_init=initializers.lecun_normal(),
        bias_init=initializers.uniform(),
        force_fp32_for_softmax=force_fp32,
        deterministic=False,
        dtype=jnp.bfloat16,
    )
    mha = nn.MultiHeadDotProductAttention(**attention_kwargs)
    init_vars = mha.init({'params': params_key, 'dropout': dropout_key}, x)
    _, updated_vars = mha.apply(
        init_vars, x, mutable=['intermediates'], sow_weights=True
    )
    self.assertEqual(
        updated_vars['intermediates']['attention_weights'][0].dtype,
        attn_weights_dtype,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from absl.testing import absltest

from flax.configurations import bool_flag, config


class MyTestCase(absltest.TestCase):
  def setUp(self):
    super().setUp()
    self.enter_context(mock.patch.object(config, '_values', {}))
    self._flag = bool_flag('test', default=False, help='Just a test flag.')

  def test_duplicate_flag(self):
    with self.assertRaisesRegex(RuntimeError, 'already defined'):
      bool_flag(self._flag.name, default=False, help='Another test flag.')

  def test_default(self):
    self.assertFalse(self._flag.value)
    self.assertFalse(config.test)

  def test_typed_update(self):
    config.update(self._flag, True)
    self.assertTrue(self._flag.value)
    self.assertTrue(config.test)

  def test_untyped_update(self):
    config.update(self._flag.name, True)
    self.assertTrue(self._flag.value)
    self.assertTrue(config.test)

  def test_update_unknown_flag(self):
    with self.assertRaisesRegex(LookupError, 'Unrecognized config option'):
      config.update('unknown', True)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tempfile import TemporaryDirectory

import jax
import jax.numpy as jnp
import orbax.checkpoint as orbax

from flax import nnx


class MLP(nnx.Module):
  def __init__(self, din: int, dmid: int, dout: int, *, rngs: nnx.Rngs):
    self.dense1 = nnx.Linear(din, dmid, rngs=rngs)
    self.dense2 = nnx.Linear(dmid, dout, rngs=rngs)

  def __call__(self, x: jax.Array) -> jax.Array:
    x = self.dense1(x)
    x = jax.nn.relu(x)
    x = self.dense2(x)
    return x


def create_model(seed: int):
  return MLP(10, 20, 30, rngs=nnx.Rngs(seed))


def create_and_save(seed: int, path: str):
  model = create_model(seed)
  state = nnx.state(model)
  # Save the parameters
  checkpointer = orbax.PyTreeCheckpointer()
  checkpointer.save(f'{path}/state', state)


def load_model(path: str) -> MLP:
  # create that model with abstract shapes
  model = nnx.eval_shape(lambda: create_model(0))
  state = nnx.state(model)
  # Load the parameters
  checkpointer = orbax.PyTreeCheckpointer()
  state = checkpointer.restore(f'{path}/state', item=state)
  # update the model with the loaded state
  nnx.update(model, state)
  return model


with TemporaryDirectory() as tmpdir:
  # create a checkpoint
  create_and_save(42, tmpdir)
  # load model from checkpoint
  model = load_model(tmpdir)
  # run the model
  y = model(jnp.ones((1, 10)))
  print(model)
  print(y)
